@@ -1,4 +1,4 @@
-/*  Copyright (C) 2018 CZ.NIC, z.s.p.o. <knot-dns@labs.nic.cz>
+/*  Copyright (C) 2020 CZ.NIC, z.s.p.o. <knot-dns@labs.nic.cz>
 
     This program is free software: you can redistribute it and/or modify
     it under the terms of the GNU General Public License as published by
@@ -11,8 +11,8 @@
     GNU General Public License for more details.
 
     You should have received a copy of the GNU General Public License
-    along with this program.  If not, see <http://www.gnu.org/licenses/>.
-*/
+    along with this program.  If not, see <https://www.gnu.org/licenses/>.
+ */
 
 #include <dlfcn.h>
 #include <fcntl.h>
@@ -135,6 +135,8 @@ int conf_mod_load_common(
 	for (module_t *mod = STATIC_MODULES; mod->api != NULL; mod++) {
 		ret = mod_load(conf, mod);
 		if (ret != KNOT_EOK) {
+			log_error("module '%s', failed to load (%s)",
+			          mod->api->name, knot_strerror(ret));
 			break;
 		}
 
@@ -343,6 +345,7 @@ void conf_mod_unload_shared(
 
 void conf_activate_modules(
 	conf_t *conf,
+	struct server *server,
 	const knot_dname_t *zone_name,
 	list_t *query_modules,
 	struct query_plan **query_plan)
@@ -393,7 +396,7 @@ void conf_activate_modules(
 		}
 
 		// Open the module.
-		knotd_mod_t *mod = query_module_open(conf, mod_id, *query_plan,
+		knotd_mod_t *mod = query_module_open(conf, server, mod_id, *query_plan,
 		                                     zone_name);
 		if (mod == NULL) {
 			MOD_ID_LOG(zone_name, error, mod_id, "failed to open");
@@ -449,7 +452,7 @@ void conf_deactivate_modules(
 	*query_plan = NULL;
 
 	// Free query modules list.
-	knotd_mod_t *mod = NULL, *next = NULL;
+	knotd_mod_t *mod, *next;
 	WALK_LIST_DELSAFE(mod, next, *query_modules) {
 		if (mod->api->unload != NULL) {
 			mod->api->unload(mod);
