@@ -1,4 +1,4 @@
-/*  Copyright (C) 2018 CZ.NIC, z.s.p.o. <knot-dns@labs.nic.cz>
+/*  Copyright (C) 2019 CZ.NIC, z.s.p.o. <knot-dns@labs.nic.cz>
 
     This program is free software: you can redistribute it and/or modify
     it under the terms of the GNU General Public License as published by
@@ -11,8 +11,8 @@
     GNU General Public License for more details.
 
     You should have received a copy of the GNU General Public License
-    along with this program.  If not, see <http://www.gnu.org/licenses/>.
-*/
+    along with this program.  If not, see <https://www.gnu.org/licenses/>.
+ */
 
 #pragma once
 
@@ -42,11 +42,11 @@ int mod_id_to_txt(
 	YP_BIN_TXT_PARAMS
 );
 
-int edns_opt_to_bin(
+int rrtype_to_bin(
 	YP_TXT_BIN_PARAMS
 );
 
-int edns_opt_to_txt(
+int rrtype_to_txt(
 	YP_BIN_TXT_PARAMS
 );
 
