@@ -1,4 +1,4 @@
-/*  Copyright (C) 2018 CZ.NIC, z.s.p.o. <knot-dns@labs.nic.cz>
+/*  Copyright (C) 2019 CZ.NIC, z.s.p.o. <knot-dns@labs.nic.cz>
 
     This program is free software: you can redistribute it and/or modify
     it under the terms of the GNU General Public License as published by
@@ -11,7 +11,7 @@
     GNU General Public License for more details.
 
     You should have received a copy of the GNU General Public License
-    along with this program.  If not, see <http://www.gnu.org/licenses/>.
+    along with this program.  If not, see <https://www.gnu.org/licenses/>.
  */
 
 #pragma once
@@ -19,9 +19,10 @@
 #include "libknot/libknot.h"
 #include "knot/server/server.h"
 
-#define CTL_FLAG_FORCE	"F"
-#define CTL_FLAG_ADD	"+"
-#define CTL_FLAG_REM	"-"
+#define CTL_FLAG_FORCE		"F"
+#define CTL_FLAG_BLOCKING	"B"
+#define CTL_FLAG_ADD		"+"
+#define CTL_FLAG_REM		"-"
 
 #define CTL_FILTER_FLUSH_OUTDIR		'd'
 
@@ -54,6 +55,7 @@ typedef enum {
 	CTL_ZONE_NOTIFY,
 	CTL_ZONE_FLUSH,
 	CTL_ZONE_SIGN,
+	CTL_ZONE_KEY_ROLL,
 	CTL_ZONE_KSK_SBM,
 	CTL_ZONE_FREEZE,
 	CTL_ZONE_THAW,
