@@ -1,4 +1,4 @@
-/*  Copyright (C) 2017 CZ.NIC, z.s.p.o. <knot-dns@labs.nic.cz>
+/*  Copyright (C) 2019 CZ.NIC, z.s.p.o. <knot-dns@labs.nic.cz>
 
     This program is free software: you can redistribute it and/or modify
     it under the terms of the GNU General Public License as published by
@@ -11,8 +11,8 @@
     GNU General Public License for more details.
 
     You should have received a copy of the GNU General Public License
-    along with this program.  If not, see <http://www.gnu.org/licenses/>.
-*/
+    along with this program.  If not, see <https://www.gnu.org/licenses/>.
+ */
 
 #pragma once
 
@@ -29,12 +29,13 @@ typedef struct {
 } knot_kasp_zone_t;
 
 int kasp_zone_load(knot_kasp_zone_t *zone,
-		   const knot_dname_t *zone_name,
-		   kasp_db_t *kdb);
+                   const knot_dname_t *zone_name,
+                   knot_lmdb_db_t *kdb,
+                   bool *kt_cfl);
 
 int kasp_zone_save(const knot_kasp_zone_t *zone,
 		   const knot_dname_t *zone_name,
-		   kasp_db_t *kdb);
+		   knot_lmdb_db_t *kdb);
 
 int kasp_zone_append(knot_kasp_zone_t *zone,
 		     const knot_kasp_key_t *appkey);
