@@ -11,7 +11,7 @@
     GNU General Public License for more details.
 
     You should have received a copy of the GNU General Public License
-    along with this program.  If not, see <http://www.gnu.org/licenses/>.
+    along with this program.  If not, see <https://www.gnu.org/licenses/>.
  */
 
 #include <assert.h>
@@ -57,7 +57,7 @@ int kdnssec_generate_key(kdnssec_ctx_t *ctx, kdnssec_generate_flags_t flags,
 	// generate key in the keystore
 
 	char *id = NULL;
-	int r = dnssec_keystore_generate_key(ctx->keystore, algorithm, size, &id);
+	int r = dnssec_keystore_generate(ctx->keystore, algorithm, size, &id);
 	if (r != KNOT_EOK) {
 		return r;
 	}
@@ -81,7 +81,7 @@ int kdnssec_generate_key(kdnssec_ctx_t *ctx, kdnssec_generate_flags_t flags,
 	dnssec_key_set_flags(dnskey, dnskey_flags(flags & DNSKEY_GENERATE_SEP_ON));
 	dnssec_key_set_algorithm(dnskey, algorithm);
 
-	r = dnssec_key_import_keystore(dnskey, ctx->keystore, id);
+	r = dnssec_keystore_export(ctx->keystore, id, dnskey);
 	if (r != KNOT_EOK) {
 		dnssec_key_free(dnskey);
 		free(id);
@@ -129,14 +129,15 @@ int kdnssec_share_key(kdnssec_ctx_t *ctx, const knot_dname_t *from_zone, const c
 		return ret;
 	}
 
-	ret = kasp_db_share_key(*ctx->kasp_db, from_zone, ctx->zone->dname, key_id);
+	ret = kasp_db_share_key(ctx->kasp_db, from_zone, ctx->zone->dname, key_id);
 	if (ret != KNOT_EOK) {
 		free(to_zone);
 		return ret;
 	}
 
 	kasp_zone_clear(ctx->zone);
-	ret = kasp_zone_load(ctx->zone, to_zone, *ctx->kasp_db);
+	ret = kasp_zone_load(ctx->zone, to_zone, ctx->kasp_db,
+	                     &ctx->keytag_conflict);
 	free(to_zone);
 	return ret;
 }
@@ -155,13 +156,13 @@ int kdnssec_delete_key(kdnssec_ctx_t *ctx, knot_kasp_key_t *key_ptr)
 	}
 
 	bool key_still_used_in_keystore = false;
-	int ret = kasp_db_delete_key(*ctx->kasp_db, ctx->zone->dname, key_ptr->id, &key_still_used_in_keystore);
+	int ret = kasp_db_delete_key(ctx->kasp_db, ctx->zone->dname, key_ptr->id, &key_still_used_in_keystore);
 	if (ret != KNOT_EOK) {
 		return ret;
 	}
 
 	if (!key_still_used_in_keystore && !key_ptr->is_pub_only) {
-		ret = dnssec_keystore_remove_key(ctx->keystore, key_ptr->id);
+		ret = dnssec_keystore_remove(ctx->keystore, key_ptr->id);
 		if (ret != KNOT_EOK) {
 			return ret;
 		}
@@ -174,10 +175,44 @@ int kdnssec_delete_key(kdnssec_ctx_t *ctx, knot_kasp_key_t *key_ptr)
 	return KNOT_EOK;
 }
 
+static bool is_published(knot_kasp_key_timing_t *timing, knot_time_t now)
+{
+	return (knot_time_cmp(timing->publish, now) <= 0 &&
+	        knot_time_cmp(timing->post_active, now) > 0 &&
+	        knot_time_cmp(timing->remove, now) > 0);
+}
+
+static bool is_ready(knot_kasp_key_timing_t *timing, knot_time_t now)
+{
+	return (knot_time_cmp(timing->ready, now) <= 0 &&
+	        knot_time_cmp(timing->active, now) > 0);
+}
+
+static bool is_active(knot_kasp_key_timing_t *timing, knot_time_t now)
+{
+	return (knot_time_cmp(timing->active, now) <= 0 &&
+	        knot_time_cmp(timing->retire, now) > 0 &&
+	        knot_time_cmp(timing->retire_active, now) > 0 &&
+	        knot_time_cmp(timing->remove, now) > 0);
+}
+
+static bool alg_has_active_zsk(kdnssec_ctx_t *ctx, uint8_t alg)
+{
+	for (size_t i = 0; i < ctx->zone->num_keys; i++) {
+		knot_kasp_key_t *k = &ctx->zone->keys[i];
+		if (dnssec_key_get_algorithm(k->key) == alg &&
+		    k->is_zsk && is_active(&k->timing, ctx->now)) {
+			return true;
+		}
+	}
+	return false;
+}
+
 /*!
  * \brief Get key feature flags from key parameters.
  */
-static void set_key(knot_kasp_key_t *kasp_key, knot_time_t now, zone_key_t *zone_key)
+static void set_key(knot_kasp_key_t *kasp_key, knot_time_t now,
+                    zone_key_t *zone_key, bool same_alg_act_zsk)
 {
 	assert(kasp_key);
 	assert(zone_key);
@@ -213,22 +248,28 @@ static void set_key(knot_kasp_key_t *kasp_key, knot_time_t now, zone_key_t *zone
 	zone_key->is_ksk = kasp_key->is_ksk;
 	zone_key->is_zsk = kasp_key->is_zsk;
 
-	zone_key->is_public = (knot_time_cmp(timing->publish, now) <= 0 &&
-	                       knot_time_cmp(timing->post_active, now) > 0 &&
-	                       knot_time_cmp(timing->remove, now) > 0);
+	zone_key->is_public = is_published(timing, now);
+	zone_key->is_ready = (zone_key->is_ksk && is_ready(timing, now));
+	zone_key->is_active = is_active(timing, now);
 
-	zone_key->is_active = (((zone_key->is_zsk && knot_time_cmp(timing->pre_active, now) <= 0) ||
-	                        (knot_time_cmp(timing->pre_active, now) <= 0 && knot_time_cmp(timing->publish, now) <= 0) ||
-	                        knot_time_cmp(timing->ready, now) <= 0 ||
-	                        knot_time_cmp(timing->active, now) <= 0) &&
-	                       knot_time_cmp(timing->retire, now) > 0 &&
-	                       (zone_key->is_zsk || knot_time_cmp(timing->post_active, now) > 0) &&
-	                       knot_time_cmp(timing->remove, now) > 0);
-
-	zone_key->cds_priority = (knot_time_cmp(timing->ready, now) <= 0 ? (
-	                          (knot_time_cmp(timing->active, now) <= 0) ? (
-	                           (knot_time_cmp(timing->retire_active, now) <= 0 ||
-	                            knot_time_cmp(timing->retire, now) <= 0) ? 0 : 1) : 2) : 0);
+	zone_key->is_ksk_active_plus = zone_key->is_ready;
+	zone_key->is_zsk_active_plus = zone_key->is_ready && !same_alg_act_zsk;
+	if (knot_time_cmp(timing->pre_active, now) <= 0 &&
+	    knot_time_cmp(timing->ready, now) > 0 &&
+	    knot_time_cmp(timing->active, now) > 0) {
+		zone_key->is_zsk_active_plus = zone_key->is_zsk;
+		zone_key->is_ksk_active_plus = (knot_time_cmp(timing->publish, now) <= 0 && zone_key->is_ksk);
+	}
+	if (knot_time_cmp(timing->retire_active, now) <= 0 &&
+	    knot_time_cmp(timing->retire, now) > 0) {
+		zone_key->is_ksk_active_plus = zone_key->is_ksk;
+		zone_key->is_zsk_active_plus = !same_alg_act_zsk;
+	} // not "else" !
+	if (knot_time_cmp(timing->post_active, now) <= 0 &&
+	    knot_time_cmp(timing->remove, now) > 0) {
+		zone_key->is_ksk_active_plus = false;
+		zone_key->is_zsk_active_plus = zone_key->is_zsk;
+	}
 }
 
 /*!
@@ -244,32 +285,10 @@ static bool is_nsec3_allowed(uint8_t algorithm)
 	}
 }
 
-static void ksk2csk(kdnssec_ctx_t *ctx, zone_keyset_t *keyset, uint8_t alg)
-{
-	for (size_t j = 0; j < keyset->count; j++) {
-		zone_key_t *key = &keyset->keys[j];
-		if (dnssec_key_get_algorithm(key->key) == alg) {
-			assert(key->is_ksk);
-			key->is_zsk = true;
-		}
-	}
-
-	for (size_t i = 0; i < ctx->zone->num_keys; i++) {
-		knot_kasp_key_t *key = &ctx->zone->keys[i];
-		if (dnssec_key_get_algorithm(key->key) == alg) {
-			assert(key->is_ksk);
-			key->is_zsk = true;
-		}
-	}
-
-	log_zone_info(ctx->zone->dname, "DNSSEC, Single-Type Signing "
-	                                "Scheme enabled");
-}
-
 static int walk_algorithms(kdnssec_ctx_t *ctx, zone_keyset_t *keyset)
 {
 	uint8_t alg_usage[256] = { 0 };
-	bool keys_changed = false, have_active_alg = false;
+	bool have_active_alg = false;
 
 	for (size_t i = 0; i < keyset->count; i++) {
 		zone_key_t *key = &keyset->keys[i];
@@ -281,14 +300,16 @@ static int walk_algorithms(kdnssec_ctx_t *ctx, zone_keyset_t *keyset)
 			                 dnssec_key_get_keytag(key->key));
 			key->is_public = false;
 			key->is_active = false;
-			key->cds_priority = 0;
+			key->is_ready = false;
+			key->is_ksk_active_plus = false;
+			key->is_zsk_active_plus = false;
 			continue;
 		}
 
 		if (key->is_ksk && key->is_public) { alg_usage[alg] |= 1; }
 		if (key->is_zsk && key->is_public) { alg_usage[alg] |= 2; }
-		if (key->is_ksk && key->is_active) { alg_usage[alg] |= 4; }
-		if (key->is_zsk && key->is_active) { alg_usage[alg] |= 8; }
+		if (key->is_ksk && (key->is_active || key->is_ksk_active_plus)) { alg_usage[alg] |= 4; }
+		if (key->is_zsk && (key->is_active || key->is_zsk_active_plus)) { alg_usage[alg] |= 8; }
 	}
 
 	for (size_t i = 0; i < sizeof(alg_usage); i++) {
@@ -296,14 +317,16 @@ static int walk_algorithms(kdnssec_ctx_t *ctx, zone_keyset_t *keyset)
 			continue; // no public keys, ignore
 		}
 		switch (alg_usage[i]) {
-		case 5: // because migrating from older version OR from manual setup
-			ksk2csk(ctx, keyset, i);
-			alg_usage[i] |= 10;
-			keys_changed = true;
-			// FALLTHROUGH
 		case 15: // all keys ready for signing
 			have_active_alg = true;
 			break;
+		case 5:
+		case 10:
+			if (ctx->policy->offline_ksk) {
+				have_active_alg = true;
+				break;
+			}
+			// else FALLTHROUGH
 		default:
 			return KNOT_DNSSEC_EMISSINGKEYTYPE;
 		}
@@ -313,10 +336,6 @@ static int walk_algorithms(kdnssec_ctx_t *ctx, zone_keyset_t *keyset)
 		return KNOT_DNSSEC_ENOKEY;
 	}
 
-	if (keys_changed) {
-		return kdnssec_ctx_commit(ctx);
-	}
-
 	return KNOT_EOK;
 }
 
@@ -329,13 +348,16 @@ static int load_private_keys(dnssec_keystore_t *keystore, zone_keyset_t *keyset)
 	assert(keyset);
 
 	for (size_t i = 0; i < keyset->count; i++) {
-		if (!keyset->keys[i].is_active) {
+		zone_key_t *key = &keyset->keys[i];
+		if (!key->is_active && !key->is_ksk_active_plus && !key->is_zsk_active_plus) {
 			continue;
 		}
-
-		zone_key_t *key = &keyset->keys[i];
-		int r = dnssec_key_import_keystore(key->key, keystore, key->id);
-		if (r != DNSSEC_EOK && r != DNSSEC_KEY_ALREADY_PRESENT) {
+		int r = dnssec_keystore_export(keystore, key->id, key->key);
+		switch (r) {
+		case DNSSEC_EOK:
+		case DNSSEC_KEY_ALREADY_PRESENT:
+			break;
+		default:
 			return r;
 		}
 	}
@@ -359,13 +381,14 @@ static void log_key_info(const zone_key_t *key, char *out, size_t out_len)
 		(void)snprintf(alg_code_str, sizeof(alg_code_str), "%d", alg_code);
 	}
 
-	(void)snprintf(out, out_len, "DNSSEC, key, tag %5d, algorithm %s%s%s%s%s",
+	(void)snprintf(out, out_len, "DNSSEC, key, tag %5d, algorithm %s%s%s%s%s%s",
 	               dnssec_key_get_keytag(key->key),
 	               (alg != NULL                ? alg->name  : alg_code_str),
 	               (key->is_ksk ? (key->is_zsk ? ", CSK" : ", KSK") : ""),
-	               (key->is_public             ? ", public" : ""),
-	               (key->cds_priority > 1      ? ", ready"  : ""),
-	               (key->is_active             ? ", active" : ""));
+	               (key->is_public             ? ", public"  : ""),
+	               (key->is_ready              ? ", ready"   : ""),
+	               (key->is_active             ? ", active"  : ""),
+	               (key->is_ksk_active_plus || key->is_zsk_active_plus ? ", active+" : ""));
 }
 
 int log_key_sort(const void *a, const void *b)
@@ -403,7 +426,9 @@ int load_zone_keys(kdnssec_ctx_t *ctx, zone_keyset_t *keyset_ptr, bool verbose)
 	char key_info[ctx->zone->num_keys][MAX_KEY_INFO];
 	for (size_t i = 0; i < ctx->zone->num_keys; i++) {
 		knot_kasp_key_t *kasp_key = &ctx->zone->keys[i];
-		set_key(kasp_key, ctx->now, &keyset.keys[i]);
+		uint8_t kk_alg = dnssec_key_get_algorithm(kasp_key->key);
+		bool same_alg_zsk = alg_has_active_zsk(ctx, kk_alg);
+		set_key(kasp_key, ctx->now, &keyset.keys[i], same_alg_zsk);
 		if (verbose) {
 			log_key_info(&keyset.keys[i], key_info[i], MAX_KEY_INFO);
 		}
