@@ -11,7 +11,7 @@
     GNU General Public License for more details.
 
     You should have received a copy of the GNU General Public License
-    along with this program.  If not, see <http://www.gnu.org/licenses/>.
+    along with this program.  If not, see <https://www.gnu.org/licenses/>.
  */
 
 #pragma once
@@ -38,7 +38,9 @@ typedef struct {
 	bool is_zsk;
 	bool is_active;
 	bool is_public;
-	int cds_priority;
+	bool is_ready;
+	bool is_zsk_active_plus;
+	bool is_ksk_active_plus;
 } zone_key_t;
 
 dynarray_declare(keyptr, zone_key_t *, DYNARRAY_VISIBILITY_PUBLIC, 1)
