@@ -11,8 +11,8 @@
     GNU General Public License for more details.
 
     You should have received a copy of the GNU General Public License
-    along with this program.  If not, see <http://www.gnu.org/licenses/>.
-*/
+    along with this program.  If not, see <https://www.gnu.org/licenses/>.
+ */
 
 #include <assert.h>
 #include <stddef.h>
@@ -28,6 +28,7 @@
 #include "knot/dnssec/policy.h"
 #include "knot/dnssec/rrset-sign.h"
 #include "knot/dnssec/zone-events.h"
+#include "knot/dnssec/zone-sign.h"
 #include "knot/nameserver/query_module.h"
 #include "knot/nameserver/process_query.h"
 
@@ -176,9 +177,8 @@ static dnssec_nsec_bitmap_t *synth_bitmap(knot_pkt_t *pkt, const knotd_qdata_t *
 	}
 
 	uint16_t qtype = knot_pkt_qtype(qdata->query);
-	bool is_apex = qdata->extra->zone
-	               && qdata->extra->zone->contents
-	               && qdata->extra->node == qdata->extra->zone->contents->apex;
+	bool is_apex = (qdata->extra->contents != NULL &&
+	                qdata->extra->node == qdata->extra->contents->apex);
 
 	bitmap_add_synth(map, is_apex);
 
@@ -308,12 +308,13 @@ static glue_t *find_glue_for(const knot_rrset_t *rr, const knot_pkt_t *pkt)
 	return NULL;
 }
 
-static bool shall_sign_rr(const knot_rrset_t *rr, const knot_pkt_t *pkt)
+static bool shall_sign_rr(const knot_rrset_t *rr, const knot_pkt_t *pkt, knotd_qdata_t *qdata)
 {
 	if (pkt->current == KNOT_ADDITIONAL) {
 		glue_t *g = find_glue_for(rr, pkt);
 		assert(g); // finds actually the node which is rr in
-		return !(g->node->flags & NODE_FLAGS_NONAUTH);
+		const zone_node_t *gn = glue_node(g, qdata->extra->node);
+		return !(gn->flags & NODE_FLAGS_NONAUTH);
 	} else {
 		return !is_deleg(pkt) || rr->type == KNOT_RRTYPE_NSEC;
 	}
@@ -337,13 +338,13 @@ static knotd_in_state_t sign_section(knotd_in_state_t state, knot_pkt_t *pkt,
 	uint16_t count_unsigned = section->count;
 	for (int i = 0; i < count_unsigned; i++) {
 		const knot_rrset_t *rr = knot_pkt_rr(section, i);
-		if (!shall_sign_rr(rr, pkt)) {
+		if (!shall_sign_rr(rr, pkt, qdata)) {
 			continue;
 		}
 
 		uint16_t rr_pos = knot_pkt_rr_offset(section, i);
 
-		uint8_t owner[KNOT_DNAME_MAXLEN] = { 0 };
+		knot_dname_storage_t owner;
 		knot_dname_unpack(owner, pkt->wire + rr_pos, sizeof(owner), pkt->wire);
 		knot_dname_to_lower(owner);
 
@@ -428,23 +429,6 @@ static knot_rrset_t *synth_dnskey(knotd_qdata_t *qdata, knotd_mod_t *mod,
 	return dnskey;
 }
 
-/* copied from the middle of zone-sign.c */
-static zone_key_t *ksk_for_cds(knotd_mod_t *mod)
-{
-	zone_key_t *res = NULL;
-	unsigned crp = mod->dnssec->policy->child_records_publish;
-	int kfc_prio = (crp == CHILD_RECORDS_ALWAYS ?
-	                0 : (crp == CHILD_RECORDS_ROLLOVER ? 1 : 2));
-	for (int i = 0; i < mod->keyset->count; i++) {
-		zone_key_t *key = &mod->keyset->keys[i];
-		if (key->is_ksk && key->cds_priority > kfc_prio) {
-			res = key;
-			kfc_prio = key->cds_priority;
-		}
-	}
-	return res;
-}
-
 static knot_rrset_t *synth_cdnskey(knotd_qdata_t *qdata, knotd_mod_t *mod,
                                    knot_mm_t *mm)
 {
@@ -458,21 +442,13 @@ static knot_rrset_t *synth_cdnskey(knotd_qdata_t *qdata, knotd_mod_t *mod,
 	dnssec_binary_t rdata = { 0 };
 	online_sign_ctx_t *ctx = knotd_mod_ctx(mod);
 	pthread_rwlock_rdlock(&ctx->signing_mutex);
-	zone_key_t *key = ksk_for_cds(mod);
-	if (key == NULL) {
-		pthread_rwlock_unlock(&ctx->signing_mutex);
-		knot_rrset_free(dnskey, mm);
-		return NULL;
+	keyptr_dynarray_t kcdnskeys = knot_zone_sign_get_cdnskeys(mod->dnssec, mod->keyset);
+	dynarray_foreach(keyptr, zone_key_t *, ksk_for_cdnskey, kcdnskeys) {
+		dnssec_key_get_rdata((*ksk_for_cdnskey)->key, &rdata);
+		assert(rdata.size > 0 && rdata.data);
+		(void)knot_rrset_add_rdata(dnskey, rdata.data, rdata.size, mm);
 	}
-	dnssec_key_get_rdata(key->key, &rdata);
-	assert(rdata.size > 0 && rdata.data);
-
-	int ret = knot_rrset_add_rdata(dnskey, rdata.data, rdata.size, mm);
 	pthread_rwlock_unlock(&ctx->signing_mutex);
-	if (ret != KNOT_EOK) {
-		knot_rrset_free(dnskey, mm);
-		return NULL;
-	}
 
 	return dnskey;
 }
@@ -490,21 +466,13 @@ static knot_rrset_t *synth_cds(knotd_qdata_t *qdata, knotd_mod_t *mod,
 	dnssec_binary_t rdata = { 0 };
 	online_sign_ctx_t *ctx = knotd_mod_ctx(mod);
 	pthread_rwlock_rdlock(&ctx->signing_mutex);
-	zone_key_t *key = ksk_for_cds(mod);
-	if (key == NULL) {
-		pthread_rwlock_unlock(&ctx->signing_mutex);
-		knot_rrset_free(ds, mm);
-		return NULL;
+	keyptr_dynarray_t kcdnskeys = knot_zone_sign_get_cdnskeys(mod->dnssec, mod->keyset);
+	dynarray_foreach(keyptr, zone_key_t *, ksk_for_cds, kcdnskeys) {
+		zone_key_calculate_ds(*ksk_for_cds, &rdata);
+		assert(rdata.size > 0 && rdata.data);
+		(void)knot_rrset_add_rdata(ds, rdata.data, rdata.size, mm);
 	}
-	zone_key_calculate_ds(key, &rdata);
-	assert(rdata.size > 0 && rdata.data);
-
-	int ret = knot_rrset_add_rdata(ds, rdata.data, rdata.size, mm);
 	pthread_rwlock_unlock(&ctx->signing_mutex);
-	if (ret != KNOT_EOK) {
-		knot_rrset_free(ds, mm);
-		return NULL;
-	}
 
 	return ds;
 }
@@ -524,7 +492,7 @@ static knotd_in_state_t pre_routine(knotd_in_state_t state, knot_pkt_t *pkt,
                                     knotd_qdata_t *qdata, knotd_mod_t *mod)
 {
 	online_sign_ctx_t *ctx = knotd_mod_ctx(mod);
-	zone_sign_reschedule_t resch = { .allow_rollover = true };
+	zone_sign_reschedule_t resch = { 0 };
 
 	(void)pkt, (void)qdata;
 
@@ -546,11 +514,14 @@ static knotd_in_state_t pre_routine(knotd_in_state_t state, knot_pkt_t *pkt,
 		}
 	}
 	if (ret == KNOT_EOK || knot_time_cmp(ctx->event_rollover, mod->dnssec->now) <= 0) {
-		update_policy_from_zone(mod->dnssec->policy, qdata->extra->zone->contents);
-		ret = knot_dnssec_key_rollover(mod->dnssec, &resch);
+		update_policy_from_zone(mod->dnssec->policy, qdata->extra->contents);
+		ret = knot_dnssec_key_rollover(mod->dnssec, KEY_ROLL_ALLOW_KSK_ROLL | KEY_ROLL_ALLOW_ZSK_ROLL, &resch);
+		if (ret != KNOT_EOK) {
+			ctx->event_rollover = knot_dnssec_failover_delay(mod->dnssec);
+		}
 	}
 	if (ret == KNOT_EOK) {
-		if (resch.plan_ds_query && mod->dnssec->policy->ksk_sbm_check_interval > 0) {
+		if (resch.plan_ds_check && mod->dnssec->policy->ksk_sbm_check_interval > 0) {
 			ctx->event_parent_ds_q = mod->dnssec->now + mod->dnssec->policy->ksk_sbm_check_interval;
 		} else {
 			ctx->event_parent_ds_q = 0;
@@ -668,19 +639,19 @@ static int online_sign_ctx_new(online_sign_ctx_t **ctx_ptr, knotd_mod_t *mod)
 		return ret;
 	}
 
-	// Force Singe-Type signing scheme. This is only important for compatibility with older versions.
-	mod->dnssec->policy->singe_type_signing = true;
+	// Historically, the default scheme is Single-Type signing.
+	if (mod->dnssec->policy->sts_default) {
+		mod->dnssec->policy->single_type_signing = true;
+	}
 
-	zone_sign_reschedule_t resch = {
-		.allow_rollover = true
-	};
-	ret = knot_dnssec_key_rollover(mod->dnssec, &resch);
+	zone_sign_reschedule_t resch = { 0 };
+	ret = knot_dnssec_key_rollover(mod->dnssec, KEY_ROLL_ALLOW_KSK_ROLL | KEY_ROLL_ALLOW_ZSK_ROLL, &resch);
 	if (ret != KNOT_EOK) {
 		free(ctx);
 		return ret;
 	}
 
-	if (resch.plan_ds_query) {
+	if (resch.plan_ds_check) {
 		ctx->event_parent_ds_q = time(NULL);
 	}
 	ctx->event_rollover = resch.next_rollover;
