@@ -1,4 +1,4 @@
-/*  Copyright (C) 2018 CZ.NIC, z.s.p.o. <knot-dns@labs.nic.cz>
+/*  Copyright (C) 2020 CZ.NIC, z.s.p.o. <knot-dns@labs.nic.cz>
 
     This program is free software: you can redistribute it and/or modify
     it under the terms of the GNU General Public License as published by
@@ -11,7 +11,7 @@
     GNU General Public License for more details.
 
     You should have received a copy of the GNU General Public License
-    along with this program.  If not, see <http://www.gnu.org/licenses/>.
+    along with this program.  If not, see <https://www.gnu.org/licenses/>.
  */
 
 #include "libknot/libknot.h"
@@ -24,7 +24,7 @@
 /*! \brief Check if given node was already visited. */
 static int wildcard_has_visited(knotd_qdata_t *qdata, const zone_node_t *node)
 {
-	struct wildcard_hit *item = NULL;
+	struct wildcard_hit *item;
 	WALK_LIST(item, qdata->extra->wildcards) {
 		if (item->node == node) {
 			return true;
@@ -116,7 +116,7 @@ static bool dname_cname_cannot_synth(const knot_rrset_t *rrset, const knot_dname
 static bool have_dnssec(knotd_qdata_t *qdata)
 {
 	return knot_pkt_has_dnssec(qdata->query) &&
-	       qdata->extra->zone->contents->dnssec;
+	       qdata->extra->contents->dnssec;
 }
 
 /*! \brief This is a wildcard-covered or any other terminal node for QNAME.
@@ -143,25 +143,20 @@ static int put_answer(knot_pkt_t *pkt, uint16_t type, knotd_qdata_t *qdata)
 
 	int ret = KNOT_EOK;
 	switch (type) {
-	case KNOT_RRTYPE_ANY: /* Append all RRSets. */ {
-		conf_val_t val = conf_zone_get(conf(), C_DISABLE_ANY,
-		                               qdata->extra->zone->name);
-		/* If ANY not allowed, set TC bit. */
-		if ((qdata->params->flags & KNOTD_QUERY_FLAG_LIMIT_ANY) &&
-		    conf_bool(&val)) {
-			knot_wire_set_tc(pkt->wire);
-			return KNOT_ESPACE;
-		}
-		for (unsigned i = 0; i < qdata->extra->node->rrset_count; ++i) {
-			rrset = node_rrset_at(qdata->extra->node, i);
-			ret = process_query_put_rr(pkt, qdata, &rrset, NULL,
+	case KNOT_RRTYPE_ANY: /* Append one or all RRSets. */
+		if (qdata->params->flags & KNOTD_QUERY_FLAG_LIMIT_ANY) {
+			rrset = node_rrset_at(qdata->extra->node, 0);
+			knot_rrset_t rrsigs = node_rrset(qdata->extra->node, KNOT_RRTYPE_RRSIG);
+			ret = process_query_put_rr(pkt, qdata, &rrset, &rrsigs,
 			                           compr_hint, put_rr_flags);
-			if (ret != KNOT_EOK) {
-				break;
+		} else {
+			for (int i = 0; i < qdata->extra->node->rrset_count && ret == KNOT_EOK; ++i) {
+				rrset = node_rrset_at(qdata->extra->node, i);
+				ret = process_query_put_rr(pkt, qdata, &rrset, NULL,
+				                           compr_hint, put_rr_flags);
 			}
 		}
 		break;
-	}
 	default: /* Single RRSet of given type. */
 		rrset = node_rrset(qdata->extra->node, type);
 		if (!knot_rrset_empty(&rrset)) {
@@ -182,7 +177,8 @@ static int put_authority_soa(knot_pkt_t *pkt, knotd_qdata_t *qdata,
 	knot_rrset_t soa = node_rrset(zone->apex, KNOT_RRTYPE_SOA);
 	knot_rrset_t rrsigs = node_rrset(zone->apex, KNOT_RRTYPE_RRSIG);
 	return process_query_put_rr(pkt, qdata, &soa, &rrsigs,
-	                            KNOT_COMPR_HINT_NONE, KNOT_PF_NOTRUNC);
+	                            KNOT_COMPR_HINT_NONE,
+	                            KNOT_PF_NOTRUNC | KNOT_PF_SOAMINTTL);
 }
 
 /*! \brief Put the delegation NS RRSet to the Authority section. */
@@ -190,7 +186,7 @@ static int put_delegation(knot_pkt_t *pkt, knotd_qdata_t *qdata)
 {
 	/* Find closest delegation point. */
 	while (!(qdata->extra->node->flags & NODE_FLAGS_DELEG)) {
-		qdata->extra->node = qdata->extra->node->parent;
+		qdata->extra->node = node_parent(qdata->extra->node);
 	}
 
 	/* Insert NS record. */
@@ -229,9 +225,10 @@ static int put_additional(knot_pkt_t *pkt, const knot_rrset_t *rr,
 
 		uint16_t hint = knot_compr_hint(info, KNOT_COMPR_HINT_RDATA +
 		                                glue->ns_pos);
-		knot_rrset_t rrsigs = node_rrset(glue->node, KNOT_RRTYPE_RRSIG);
+		const zone_node_t *gluenode = glue_node(glue, qdata->extra->node);
+		knot_rrset_t rrsigs = node_rrset(gluenode, KNOT_RRTYPE_RRSIG);
 		for (int k = 0; k < ar_type_count; ++k) {
-			knot_rrset_t rrset = node_rrset(glue->node, ar_type_list[k]);
+			knot_rrset_t rrset = node_rrset(gluenode, ar_type_list[k]);
 			if (knot_rrset_empty(&rrset)) {
 				continue;
 			}
@@ -278,8 +275,8 @@ static int follow_cname(knot_pkt_t *pkt, uint16_t rrtype, knotd_qdata_t *qdata)
 			qdata->rcode = KNOT_RCODE_YXDOMAIN;
 		} else {
 			knot_rrset_t dname_rr = cname_rr;
-			int ret = dname_cname_synth(&dname_rr, qdata->name,
-			                            &cname_rr, &pkt->mm);
+			ret = dname_cname_synth(&dname_rr, qdata->name,
+			                        &cname_rr, &pkt->mm);
 			if (ret != KNOT_EOK) {
 				qdata->rcode = KNOT_RCODE_SERVFAIL;
 				return KNOTD_IN_STATE_ERROR;
@@ -359,7 +356,7 @@ static int name_not_found(knot_pkt_t *pkt, knotd_qdata_t *qdata)
 		/* Find wildcard child in the zone. */
 		const zone_node_t *wildcard_node =
 			zone_contents_find_wildcard_child(
-				qdata->extra->zone->contents, qdata->extra->encloser);
+				qdata->extra->contents, qdata->extra->encloser);
 
 		qdata->extra->node = wildcard_node;
 		assert(qdata->extra->node != NULL);
@@ -385,7 +382,7 @@ static int name_not_found(knot_pkt_t *pkt, knotd_qdata_t *qdata)
 	/* Look up an authoritative encloser or its parent. */
 	const zone_node_t *node = qdata->extra->encloser;
 	while (node->rrset_count == 0 || node->flags & NODE_FLAGS_NONAUTH) {
-		node = node->parent;
+		node = node_parent(node);
 		assert(node);
 	}
 
@@ -400,7 +397,7 @@ static int name_not_found(knot_pkt_t *pkt, knotd_qdata_t *qdata)
 
 static int solve_name(int state, knot_pkt_t *pkt, knotd_qdata_t *qdata)
 {
-	int ret = zone_contents_find_dname(qdata->extra->zone->contents, qdata->name,
+	int ret = zone_contents_find_dname(qdata->extra->contents, qdata->name,
 	                                   &qdata->extra->node, &qdata->extra->encloser,
 	                                   &qdata->extra->previous);
 
@@ -455,7 +452,7 @@ static int solve_answer_dnssec(int state, knot_pkt_t *pkt, knotd_qdata_t *qdata,
 static int solve_authority(int state, knot_pkt_t *pkt, knotd_qdata_t *qdata, void *ctx)
 {
 	int ret = KNOT_ERROR;
-	const zone_contents_t *zone_contents = qdata->extra->zone->contents;
+	const zone_contents_t *zone_contents = qdata->extra->contents;
 
 	switch (state) {
 	case KNOTD_IN_STATE_HIT:    /* Positive response. */
@@ -581,7 +578,7 @@ static int answer_query(knot_pkt_t *pkt, knotd_qdata_t *qdata)
 {
 	int state = KNOTD_IN_STATE_BEGIN;
 	struct query_plan *plan = qdata->extra->zone->query_plan;
-	struct query_step *step = NULL;
+	struct query_step *step;
 
 	bool with_dnssec = have_dnssec(qdata);
 
@@ -646,7 +643,7 @@ int internet_process_query(knot_pkt_t *pkt, knotd_qdata_t *qdata)
 	/* No applicable ACL, refuse transaction security. */
 	if (knot_pkt_has_tsig(qdata->query)) {
 		/* We have been challenged... */
-		NS_NEED_AUTH(qdata, qdata->extra->zone->name, ACL_ACTION_NONE);
+		NS_NEED_AUTH(qdata, ACL_ACTION_NONE);
 
 		/* Reserve space for TSIG. */
 		int ret = knot_pkt_reserve(pkt, knot_tsig_wire_size(&qdata->sign.tsig_key));
