@@ -1,4 +1,4 @@
-/*  Copyright (C) 2018 CZ.NIC, z.s.p.o. <knot-dns@labs.nic.cz>
+/*  Copyright (C) 2020 CZ.NIC, z.s.p.o. <knot-dns@labs.nic.cz>
 
     This program is free software: you can redistribute it and/or modify
     it under the terms of the GNU General Public License as published by
@@ -11,7 +11,7 @@
     GNU General Public License for more details.
 
     You should have received a copy of the GNU General Public License
-    along with this program.  If not, see <http://www.gnu.org/licenses/>.
+    along with this program.  If not, see <https://www.gnu.org/licenses/>.
  */
 
 #include <urcu.h>
@@ -238,13 +238,13 @@ static int answer_edns_init(const knot_pkt_t *query, knot_pkt_t *resp,
 	}
 
 	/* Initialize OPT record. */
-	int16_t max_payload;
+	uint16_t max_payload;
 	switch (qdata->params->remote->ss_family) {
 	case AF_INET:
-		max_payload = conf()->cache.srv_max_ipv4_udp_payload;
+		max_payload = conf()->cache.srv_udp_max_payload_ipv4;
 		break;
 	case AF_INET6:
-		max_payload = conf()->cache.srv_max_ipv6_udp_payload;
+		max_payload = conf()->cache.srv_udp_max_payload_ipv6;
 		break;
 	default:
 		return KNOT_ERROR;
@@ -292,7 +292,7 @@ static int answer_edns_init(const knot_pkt_t *query, knot_pkt_t *resp,
 	}
 
 	/* Initialize EDNS Client Subnet if configured and present in query. */
-	if (conf()->cache.use_ecs) {
+	if (conf()->cache.srv_ecs) {
 		uint8_t *ecs_opt = knot_pkt_edns_option(query, KNOT_EDNS_OPTION_CLIENT_SUBNET);
 		if (ecs_opt != NULL) {
 			qdata->ecs = mm_alloc(qdata->mm, sizeof(knot_edns_client_subnet_t));
@@ -331,7 +331,7 @@ static int answer_edns_put(knot_pkt_t *resp, knotd_qdata_t *qdata)
 	/* Add ECS if present. */
 	int ret = KNOT_EOK;
 	if (qdata->ecs != NULL) {
-		uint8_t *ecs_opt = knot_edns_get_option(&qdata->opt_rr, KNOT_EDNS_OPTION_CLIENT_SUBNET);
+		uint8_t *ecs_opt = knot_edns_get_option(&qdata->opt_rr, KNOT_EDNS_OPTION_CLIENT_SUBNET, NULL);
 		if (ecs_opt != NULL) {
 			uint8_t *ecs_data = knot_edns_opt_get_data(ecs_opt);
 			uint16_t ecs_len = knot_edns_opt_get_length(ecs_opt);
@@ -385,30 +385,30 @@ static int prepare_answer(knot_pkt_t *query, knot_pkt_t *resp, knot_layer_t *ctx
 	if (has_limit) {
 		resp->max_size = KNOT_WIRE_MIN_PKTSIZE;
 		if (knot_pkt_has_edns(query)) {
-			uint16_t server;
+			uint16_t server_size;
 			switch (qdata->params->remote->ss_family) {
 			case AF_INET:
-				server = conf()->cache.srv_max_ipv4_udp_payload;
+				server_size = conf()->cache.srv_udp_max_payload_ipv4;
 				break;
 			case AF_INET6:
-				server = conf()->cache.srv_max_ipv6_udp_payload;
+				server_size = conf()->cache.srv_udp_max_payload_ipv6;
 				break;
 			default:
 				return KNOT_ERROR;
 			}
-			uint16_t client = knot_edns_get_payload(query->opt_rr);
-			uint16_t transfer = MIN(client, server);
+			uint16_t client_size = knot_edns_get_payload(query->opt_rr);
+			uint16_t transfer = MIN(client_size, server_size);
 			resp->max_size = MAX(resp->max_size, transfer);
 		}
 	} else {
 		resp->max_size = KNOT_WIRE_MAX_PKTSIZE;
 	}
 
-	/* Query MUST carry a question. */
+	/* All supported OPCODEs require a question. */
 	const knot_dname_t *qname = knot_pkt_qname(query);
 	if (qname == NULL) {
-		qdata->rcode = KNOT_RCODE_FORMERR;
-		return KNOT_EMALF;
+		qdata->rcode = KNOT_RCODE_NOTIMPL;
+		return KNOT_ENOTSUP;
 	}
 
 	/* Convert query QNAME to lowercase, but keep original QNAME case.
@@ -419,6 +419,9 @@ static int prepare_answer(knot_pkt_t *query, knot_pkt_t *resp, knot_layer_t *ctx
 
 	/* Find zone for QNAME. */
 	qdata->extra->zone = answer_zone_find(query, server->zone_db);
+	if (qdata->extra->zone != NULL && qdata->extra->contents == NULL) {
+		qdata->extra->contents = qdata->extra->zone->contents;
+	}
 
 	return KNOT_EOK;
 }
@@ -453,6 +456,7 @@ static int process_query_err(knot_layer_t *ctx, knot_pkt_t *pkt)
 
 	/* Set TC bit if required. */
 	if (qdata->err_truncated) {
+		knot_wire_set_aa(pkt->wire);
 		knot_wire_set_tc(pkt->wire);
 	}
 
@@ -471,7 +475,7 @@ static int process_query_err(knot_layer_t *ctx, knot_pkt_t *pkt)
 	}
 
 	/* Set final RCODE to packet. */
-	if (qdata->rcode == KNOT_RCODE_NOERROR) {
+	if (qdata->rcode == KNOT_RCODE_NOERROR && !qdata->err_truncated) {
 		/* Default RCODE is SERVFAIL if not otherwise specified. */
 		qdata->rcode = KNOT_RCODE_SERVFAIL;
 	}
@@ -514,7 +518,7 @@ static int process_query_out(knot_layer_t *ctx, knot_pkt_t *pkt)
 	knotd_qdata_t *qdata = QUERY_DATA(ctx);
 	struct query_plan *plan = conf()->query_plan;
 	struct query_plan *zone_plan = NULL;
-	struct query_step *step = NULL;
+	struct query_step *step;
 
 	int next_state = KNOT_STATE_PRODUCE;
 
@@ -606,9 +610,10 @@ finish:
 	return next_state;
 }
 
-bool process_query_acl_check(conf_t *conf, const knot_dname_t *zone_name,
-                             acl_action_t action, knotd_qdata_t *qdata)
+bool process_query_acl_check(conf_t *conf, acl_action_t action,
+                             knotd_qdata_t *qdata)
 {
+	const knot_dname_t *zone_name = qdata->extra->zone->name;
 	knot_pkt_t *query = qdata->query;
 	const struct sockaddr_storage *query_source = qdata->params->remote;
 	knot_tsig_key_t tsig = { 0 };
@@ -626,9 +631,9 @@ bool process_query_acl_check(conf_t *conf, const knot_dname_t *zone_name,
 
 	/* Check if authenticated. */
 	conf_val_t acl = conf_zone_get(conf, C_ACL, zone_name);
-	if (!acl_allowed(conf, &acl, action, query_source, &tsig)) {
+	if (!acl_allowed(conf, &acl, action, query_source, &tsig, zone_name, query)) {
 		char addr_str[SOCKADDR_STRLEN] = { 0 };
-		sockaddr_tostr(addr_str, sizeof(addr_str), (struct sockaddr *)query_source);
+		sockaddr_tostr(addr_str, sizeof(addr_str), query_source);
 		const knot_lookup_t *act = knot_lookup_by_id((knot_lookup_t *)acl_actions,
 		                                             action);
 		char *key_name = knot_dname_to_str_alloc(tsig.name);
@@ -782,7 +787,7 @@ fail:
 /*! \brief Synthesize RRSIG for given parameters, store in 'qdata' for later use */
 static int put_rrsig(const knot_dname_t *sig_owner, uint16_t type,
                      const knot_rrset_t *rrsigs, knot_rrinfo_t *rrinfo,
-                     knotd_qdata_t *qdata)
+                     uint32_t ttl_limit, knotd_qdata_t *qdata)
 {
 	knot_rdataset_t synth_rrs;
 	knot_rdataset_init(&synth_rrs);
@@ -809,8 +814,9 @@ static int put_rrsig(const knot_dname_t *sig_owner, uint16_t type,
 		knot_rdataset_clear(&synth_rrs, qdata->mm);
 		return KNOT_ENOMEM;
 	}
+	uint32_t orig_ttl = knot_rrsig_original_ttl(synth_rrs.rdata);
 	knot_rrset_init(&info->synth_rrsig, owner_copy, rrsigs->type,
-	                rrsigs->rclass, rrsigs->ttl);
+	                rrsigs->rclass, MIN(orig_ttl, ttl_limit));
 	/* Store filtered signature. */
 	info->synth_rrsig.rrs = synth_rrs;
 
@@ -866,12 +872,17 @@ int process_query_put_rr(knot_pkt_t *pkt, knotd_qdata_t *qdata,
 		return ret;
 	}
 
+	uint32_t rrsig_ttl_limit = UINT32_MAX;
+	if ((flags & KNOT_PF_SOAMINTTL) && to_add.type == KNOT_RRTYPE_SOA) {
+		rrsig_ttl_limit = knot_soa_minimum(to_add.rrs.rdata);
+	}
+
 	const bool inserted = (prev_count != pkt->rrset_count);
 	if (inserted &&
 	    !knot_rrset_empty(rrsigs) && rr->type != KNOT_RRTYPE_RRSIG) {
 		// Get rrinfo of just inserted RR.
 		knot_rrinfo_t *rrinfo = &pkt->rr_info[pkt->rrset_count - 1];
-		ret = put_rrsig(rr->owner, rr->type, rrsigs, rrinfo, qdata);
+		ret = put_rrsig(rr->owner, rr->type, rrsigs, rrinfo, rrsig_ttl_limit, qdata);
 	}
 
 	return ret;
