@@ -1,4 +1,4 @@
-/*  Copyright (C) 2018 CZ.NIC, z.s.p.o. <knot-dns@labs.nic.cz>
+/*  Copyright (C) 2019 CZ.NIC, z.s.p.o. <knot-dns@labs.nic.cz>
 
     This program is free software: you can redistribute it and/or modify
     it under the terms of the GNU General Public License as published by
@@ -11,7 +11,7 @@
     GNU General Public License for more details.
 
     You should have received a copy of the GNU General Public License
-    along with this program.  If not, see <http://www.gnu.org/licenses/>.
+    along with this program.  If not, see <https://www.gnu.org/licenses/>.
  */
 
 #pragma once
@@ -27,6 +27,7 @@ const knot_layer_api_t *process_query_layer(void);
 /*! \brief Query processing intermediate data. */
 typedef struct knotd_qdata_extra {
 	const zone_t *zone;  /*!< Zone from which is answered. */
+	const zone_contents_t *contents; /*!< Zone contents from which is answered. */
 	list_t wildcards;    /*!< Visited wildcards. */
 	list_t rrsigs;       /*!< Section RRSIGs. */
 	uint8_t *opt_rr_pos; /*!< Place of the OPT RR in wire. */
@@ -35,7 +36,7 @@ typedef struct knotd_qdata_extra {
 	const zone_node_t *node, *encloser, *previous;
 
 	/* Original QNAME case. */
-	uint8_t orig_qname[KNOT_DNAME_MAXLEN];
+	knot_dname_storage_t orig_qname;
 
 	/* Extensions. */
 	void *ext;
@@ -61,13 +62,12 @@ struct rrsig_info {
  * \brief Check current query against ACL.
  *
  * \param conf       Configuration.
- * \param zone_name  Current zone name.
  * \param action     ACL action.
  * \param qdata      Query data.
  * \return true if accepted, false if denied.
  */
-bool process_query_acl_check(conf_t *conf, const knot_dname_t *zone_name,
-                             acl_action_t action, knotd_qdata_t *qdata);
+bool process_query_acl_check(conf_t *conf, acl_action_t action,
+                             knotd_qdata_t *qdata);
 
 /*!
  * \brief Verify current query transaction security and update query data.
