@@ -1,4 +1,4 @@
-/*  Copyright (C) 2019 CZ.NIC, z.s.p.o. <knot-dns@labs.nic.cz>
+/*  Copyright (C) 2020 CZ.NIC, z.s.p.o. <knot-dns@labs.nic.cz>
 
     This program is free software: you can redistribute it and/or modify
     it under the terms of the GNU General Public License as published by
@@ -11,14 +11,13 @@
     GNU General Public License for more details.
 
     You should have received a copy of the GNU General Public License
-    along with this program.  If not, see <http://www.gnu.org/licenses/>.
+    along with this program.  If not, see <https://www.gnu.org/licenses/>.
  */
 
 #define __APPLE_USE_RFC_3542
 
 #include <stdlib.h>
 #include <assert.h>
-#include <urcu.h>
 #include <netinet/tcp.h>
 
 #include "libknot/errcode.h"
@@ -29,6 +28,7 @@
 #include "knot/conf/migration.h"
 #include "knot/conf/module.h"
 #include "knot/dnssec/kasp/kasp_db.h"
+#include "knot/journal/journal_basic.h"
 #include "knot/server/server.h"
 #include "knot/server/udp-handler.h"
 #include "knot/server/tcp-handler.h"
@@ -50,31 +50,41 @@ enum {
 /*! \brief Unbind interface and clear the structure. */
 static void server_deinit_iface(iface_t *iface)
 {
+	assert(iface);
+
 	/* Free UDP handler. */
-	for (int i = 0; i < iface->fd_udp_count; i++) {
-		if (iface->fd_udp[i] > -1) {
-			close(iface->fd_udp[i]);
+	if (iface->fd_udp != NULL) {
+		for (int i = 0; i < iface->fd_udp_count; i++) {
+			if (iface->fd_udp[i] > -1) {
+				close(iface->fd_udp[i]);
+			}
 		}
+		free(iface->fd_udp);
 	}
-	free(iface->fd_udp);
 
 	/* Free TCP handler. */
-	if (iface->fd_tcp > -1) {
-		close(iface->fd_tcp);
+	if (iface->fd_tcp != NULL) {
+		for (int i = 0; i < iface->fd_tcp_count; i++) {
+			if (iface->fd_tcp[i] > -1) {
+				close(iface->fd_tcp[i]);
+			}
+		}
+		free(iface->fd_tcp);
 	}
 
-	memset(iface, 0, sizeof(*iface));
+	free(iface);
 }
 
-/*! \brief Unbind and dispose given interface. */
-static void server_remove_iface(iface_t *iface)
+/*! \brief Deinit server interface list. */
+static void server_deinit_iface_list(list_t *ifaces)
 {
-	if (!iface) {
-		return;
+	if (ifaces != NULL) {
+		iface_t *iface, *next;
+		WALK_LIST_DELSAFE(iface, next, *ifaces) {
+			server_deinit_iface(iface);
+		}
+		free(ifaces);
 	}
-
-	server_deinit_iface(iface);
-	free(iface);
 }
 
 /*! \brief Set lower bound for socket option. */
@@ -135,7 +145,32 @@ static bool enable_pktinfo(int sock, int family)
 	return setsockopt(sock, level, option, &on, sizeof(on)) == 0;
 }
 
-/**
+/*!
+ * Linux 3.15 has IP_PMTUDISC_OMIT which makes sockets
+ * ignore PMTU information and send packets with DF=0.
+ * Fragmentation is allowed if and only if the packet
+ * size exceeds the outgoing interface MTU or the packet
+ * encounters smaller MTU link in network.
+ * This mitigates DNS fragmentation attacks by preventing
+ * forged PMTU information.
+ * FreeBSD already has same semantics without setting
+ * the option.
+ */
+static int disable_pmtudisc(int sock, int family)
+{
+#if defined(IP_MTU_DISCOVER) && defined(IP_PMTUDISC_OMIT)
+	if (family == AF_INET) {
+		int action_omit = IP_PMTUDISC_OMIT;
+		if (setsockopt(sock, IPPROTO_IP, IP_MTU_DISCOVER, &action_omit,
+		    sizeof(action_omit)) != 0) {
+			return knot_map_errno();
+		}
+	}
+#endif
+	return KNOT_EOK;
+}
+
+/*!
  * \brief Enable TCP Fast Open.
  */
 static int enable_fastopen(int sock, int backlog)
@@ -154,247 +189,211 @@ static int enable_fastopen(int sock, int backlog)
 }
 
 /*!
- * \brief Initialize new interface from config value.
+ * \brief Create and initialize new interface.
  *
  * Both TCP and UDP sockets will be created for the interface.
  *
- * \param new_if Allocated memory for the interface.
- * \param cfg_if Interface template from config.
+ * \param addr              Socket address.
+ * \param udp_thread_count  Number of created UDP workers.
+ * \param tcp_thread_count  Number of created TCP workers.
+ * \param tcp_reuseport     Indication if reuseport on TCP is enabled.
  *
- * \retval 0 if successful (EOK).
- * \retval <0 on errors (EACCES, EINVAL, ENOMEM, EADDRINUSE).
+ * \retval Pointer to a new initialized inteface.
+ * \retval NULL if error.
  */
-static int server_init_iface(iface_t *new_if, struct sockaddr_storage *addr, int udp_thread_count)
+static iface_t *server_init_iface(struct sockaddr_storage *addr,
+                                  int udp_thread_count, int tcp_thread_count,
+                                  bool tcp_reuseport)
 {
-	/* Initialize interface. */
-	int ret = 0;
-	memset(new_if, 0, sizeof(iface_t));
-	memcpy(&new_if->addr, addr, sizeof(struct sockaddr_storage));
+	iface_t *new_if = calloc(1, sizeof(*new_if));
+	if (new_if == NULL) {
+		log_error("failed to initialize interface");
+		return NULL;
+	}
+	memcpy(&new_if->addr, addr, sizeof(*addr));
 
 	/* Convert to string address format. */
 	char addr_str[SOCKADDR_STRLEN] = { 0 };
-	sockaddr_tostr(addr_str, sizeof(addr_str), (struct sockaddr *)addr);
+	sockaddr_tostr(addr_str, sizeof(addr_str), addr);
 
 	int udp_socket_count = 1;
 	int udp_bind_flags = 0;
+	int tcp_socket_count = 1;
+	int tcp_bind_flags = 0;
 
 #ifdef ENABLE_REUSEPORT
 	udp_socket_count = udp_thread_count;
 	udp_bind_flags |= NET_BIND_MULTIPLE;
+
+	if (tcp_reuseport) {
+		tcp_socket_count = tcp_thread_count;
+		tcp_bind_flags |= NET_BIND_MULTIPLE;
+	}
 #endif
 
 	new_if->fd_udp = malloc(udp_socket_count * sizeof(int));
-	if (!new_if->fd_udp) {
-		return KNOT_ENOMEM;
+	new_if->fd_tcp = malloc(tcp_socket_count * sizeof(int));
+	if (new_if->fd_udp == NULL || new_if->fd_tcp == NULL) {
+		log_error("failed to initialize interface");
+		server_deinit_iface(new_if);
+		return NULL;
 	}
 
-	/* Initialize the sockets to ensure safe early deinitialization. */
-	for (int i = 0; i < udp_socket_count; i++) {
-		new_if->fd_udp[i] = -1;
-	}
-	new_if->fd_tcp = -1;
-
-	bool warn_bind = false;
-	bool warn_bufsize = false;
+	bool warn_bind = true;
+	bool warn_bufsize = true;
+	bool warn_pktinfo = true;
+	bool warn_flag_misc = true;
 
 	/* Create bound UDP sockets. */
-	for (int i = 0; i < udp_socket_count; i++ ) {
-		int sock = net_bound_socket(SOCK_DGRAM, (struct sockaddr *)addr, udp_bind_flags);
+	for (int i = 0; i < udp_socket_count; i++) {
+		int sock = net_bound_socket(SOCK_DGRAM, addr, udp_bind_flags);
 		if (sock == KNOT_EADDRNOTAVAIL) {
 			udp_bind_flags |= NET_BIND_NONLOCAL;
-			sock = net_bound_socket(SOCK_DGRAM, (struct sockaddr *)addr, udp_bind_flags);
-			if (sock >= 0 && !warn_bind) {
+			sock = net_bound_socket(SOCK_DGRAM, addr, udp_bind_flags);
+			if (sock >= 0 && warn_bind) {
 				log_warning("address %s UDP bound, but required nonlocal bind", addr_str);
-				warn_bind = true;
+				warn_bind = false;
 			}
 		}
 
 		if (sock < 0) {
-			log_error("cannot bind address %s (%s)", addr_str,
+			log_error("cannot bind address %s UDP (%s)", addr_str,
 			          knot_strerror(sock));
 			server_deinit_iface(new_if);
-			return sock;
+			return NULL;
 		}
 
 		if (!enlarge_net_buffers(sock, UDP_MIN_RCVSIZE, UDP_MIN_SNDSIZE) &&
-		    !warn_bufsize) {
+		    warn_bufsize) {
 			log_warning("failed to set network buffer sizes for UDP");
-			warn_bufsize = true;
+			warn_bufsize = false;
 		}
 
-		if (sockaddr_is_any((struct sockaddr *)addr) && !enable_pktinfo(sock, addr->ss_family)) {
+		if (sockaddr_is_any(addr) && !enable_pktinfo(sock, addr->ss_family) &&
+		    warn_pktinfo) {
 			log_warning("failed to enable received packet information retrieval");
+			warn_pktinfo = false;
+		}
+
+		int ret = disable_pmtudisc(sock, addr->ss_family);
+		if (ret != KNOT_EOK && warn_flag_misc) {
+			log_warning("failed to disable Path MTU discovery for IPv4/UDP (%s)",
+			            knot_strerror(ret));
+			warn_flag_misc = false;
 		}
 
 		new_if->fd_udp[new_if->fd_udp_count] = sock;
 		new_if->fd_udp_count += 1;
 	}
 
-	/* Create bound TCP socket. */
-	int tcp_bind_flags = 0;
-	int sock = net_bound_socket(SOCK_STREAM, (struct sockaddr *)addr, tcp_bind_flags);
-	if (sock == KNOT_EADDRNOTAVAIL) {
-		tcp_bind_flags |= NET_BIND_NONLOCAL;
-		sock = net_bound_socket(SOCK_STREAM, (struct sockaddr *)addr, tcp_bind_flags);
-		if (sock >= 0) {
-			log_warning("address %s TCP bound, but required nonlocal bind", addr_str);
+	warn_bind = true;
+	warn_bufsize = true;
+	warn_flag_misc = true;
+
+	/* Create bound TCP sockets. */
+	for (int i = 0; i < tcp_socket_count; i++) {
+		int sock = net_bound_socket(SOCK_STREAM, addr, tcp_bind_flags);
+		if (sock == KNOT_EADDRNOTAVAIL) {
+			tcp_bind_flags |= NET_BIND_NONLOCAL;
+			sock = net_bound_socket(SOCK_STREAM, addr, tcp_bind_flags);
+			if (sock >= 0 && warn_bind) {
+				log_warning("address %s TCP bound, but required nonlocal bind", addr_str);
+				warn_bind = false;
+			}
+		}
+
+		if (sock < 0) {
+			log_error("cannot bind address %s TCP (%s)", addr_str,
+			          knot_strerror(sock));
+			server_deinit_iface(new_if);
+			return NULL;
+		}
+
+		if (!enlarge_net_buffers(sock, TCP_MIN_RCVSIZE, TCP_MIN_SNDSIZE) &&
+		    warn_bufsize) {
+			log_warning("failed to set network buffer sizes for TCP");
+			warn_bufsize = false;
+		}
+
+		new_if->fd_tcp[new_if->fd_tcp_count] = sock;
+		new_if->fd_tcp_count += 1;
+
+		/* Listen for incoming connections. */
+		int ret = listen(sock, TCP_BACKLOG_SIZE);
+		if (ret < 0) {
+			log_error("failed to listen on TCP interface %s", addr_str);
+			server_deinit_iface(new_if);
+			return NULL;
+		}
+
+		/* TCP Fast Open. */
+		ret = enable_fastopen(sock, TCP_BACKLOG_SIZE);
+		if (ret < 0 && warn_flag_misc) {
+			log_warning("failed to enable TCP Fast Open on %s (%s)",
+			            addr_str, knot_strerror(ret));
+			warn_flag_misc = false;
 		}
 	}
 
-	if (sock < 0) {
-		log_error("cannot bind address %s (%s)", addr_str,
-		          knot_strerror(sock));
-		server_deinit_iface(new_if);
-		return sock;
-	}
-
-	if (!enlarge_net_buffers(sock, TCP_MIN_RCVSIZE, TCP_MIN_SNDSIZE)) {
-		log_warning("failed to set network buffer sizes for TCP");
-	}
-
-	new_if->fd_tcp = sock;
-
-	/* Listen for incoming connections. */
-	ret = listen(sock, TCP_BACKLOG_SIZE);
-	if (ret < 0) {
-		log_error("failed to listen on TCP interface %s", addr_str);
-		server_deinit_iface(new_if);
-		return KNOT_ERROR;
-	}
-
-	/* TCP Fast Open. */
-	ret = enable_fastopen(sock, TCP_BACKLOG_SIZE);
-	if (ret < 0) {
-		log_warning("failed to enable TCP Fast Open on %s (%s)",
-		            addr_str, knot_strerror(ret));
-	}
-
-	return KNOT_EOK;
+	return new_if;
 }
 
-static void remove_ifacelist(struct ref *p)
+/*! \brief Initialize bound sockets according to configuration. */
+static int configure_sockets(conf_t *conf, server_t *s)
 {
-	ifacelist_t *ifaces = (ifacelist_t *)p;
-
-	/* Remove deprecated interfaces. */
-	char addr_str[SOCKADDR_STRLEN] = {0};
-	iface_t *n = NULL, *m = NULL;
-	WALK_LIST_DELSAFE(n, m, ifaces->u) {
-		sockaddr_tostr(addr_str, sizeof(addr_str), (struct sockaddr *)&n->addr);
-		log_info("removing interface %s", addr_str);
-		server_remove_iface(n);
-	}
-	WALK_LIST_DELSAFE(n, m, ifaces->l) {
-		free(n);
+	if (s->state & ServerRunning) {
+		return KNOT_EOK;
 	}
 
-	free(ifaces);
-}
-
-/*!
- * \brief Update bound sockets according to configuration.
- *
- * \param server Server instance.
- * \return number of added sockets.
- */
-static int reconfigure_sockets(conf_t *conf, server_t *s)
-{
 	/* Prepare helper lists. */
-	int bound = 0;
-	ifacelist_t *oldlist = s->ifaces;
-	ifacelist_t *newlist = malloc(sizeof(ifacelist_t));
-	ref_init(&newlist->ref, &remove_ifacelist);
-	ref_retain(&newlist->ref);
-	init_list(&newlist->u);
-	init_list(&newlist->l);
-
-	/* Duplicate current list. */
-	/*! \note Pointers to addr, handlers etc. will be shared. */
-	if (s->ifaces) {
-		list_dup(&s->ifaces->u, &s->ifaces->l, sizeof(iface_t));
+	list_t *newlist = malloc(sizeof(list_t));
+	if (newlist == NULL) {
+		return KNOT_ENOMEM;
 	}
+	init_list(newlist);
+
+#ifdef ENABLE_REUSEPORT
+	/* Log info if reuseport is used and for what protocols. */
+	log_info("using reuseport for UDP%s", conf->cache.srv_tcp_reuseport ? " and TCP" : "");
+#endif
 
 	/* Update bound interfaces. */
 	conf_val_t listen_val = conf_get(conf, C_SRV, C_LISTEN);
 	conf_val_t rundir_val = conf_get(conf, C_SRV, C_RUNDIR);
 	char *rundir = conf_abs_path(&rundir_val, NULL);
 	while (listen_val.code == KNOT_EOK) {
-		iface_t *m = NULL;
-
-		/* Find already matching interface. */
-		int found_match = 0;
+		/* Log interface binding. */
 		struct sockaddr_storage addr = conf_addr(&listen_val, rundir);
-		if (s->ifaces) {
-			WALK_LIST(m, s->ifaces->u) {
-				/* Matching port and address. */
-				if (sockaddr_cmp((struct sockaddr *)&addr,
-				                 (struct sockaddr *)&m->addr) == 0) {
-					found_match = 1;
-					break;
-				}
-			}
-		}
+		char addr_str[SOCKADDR_STRLEN] = { 0 };
+		sockaddr_tostr(addr_str, sizeof(addr_str), &addr);
+		log_info("binding to interface %s", addr_str);
 
-		/* Found already bound interface. */
-		if (found_match) {
-			rem_node((node_t *)m);
-		} else {
-			char addr_str[SOCKADDR_STRLEN] = { 0 };
-			sockaddr_tostr(addr_str, sizeof(addr_str), (struct sockaddr *)&addr);
-			log_info("binding to interface %s", addr_str);
-
-			/* Create new interface. */
-			m = malloc(sizeof(iface_t));
-			unsigned size = s->handlers[IO_UDP].handler.unit->size;
-			if (server_init_iface(m, &addr, size) < 0) {
-				free(m);
-				m = 0;
-			}
-		}
-
-		/* Move to new list. */
-		if (m) {
-			add_tail(&newlist->l, (node_t *)m);
-			++bound;
+		/* Create new interface. */
+		unsigned size_udp = s->handlers[IO_UDP].handler.unit->size;
+		unsigned size_tcp = s->handlers[IO_TCP].handler.unit->size;
+		bool tcp_reuseport = conf->cache.srv_tcp_reuseport;
+		iface_t *new_if = server_init_iface(&addr, size_udp, size_tcp, tcp_reuseport);
+		if (new_if != NULL) {
+			add_tail(newlist, &new_if->n);
 		}
 
 		conf_val_next(&listen_val);
 	}
 	free(rundir);
 
-	/* Wait for readers that are reconfiguring right now. */
-	/*! \note This subsystem will be reworked in #239 */
-	for (unsigned proto = IO_UDP; proto <= IO_TCP; ++proto) {
-		dt_unit_t *tu = s->handlers[proto].handler.unit;
-		iohandler_t *ioh = &s->handlers[proto].handler;
-		for (unsigned i = 0; i < tu->size; ++i) {
-			while (ioh->thread_state[i] & ServerReload) {
-				sleep(1);
-			}
-		}
-	}
-
 	/* Publish new list. */
 	s->ifaces = newlist;
 
-	/* Update TCP+UDP ifacelist (reload all threads). */
+	/* Set the ID's (thread_id) of both the TCP and UDP threads. */
 	unsigned thread_count = 0;
 	for (unsigned proto = IO_UDP; proto <= IO_TCP; ++proto) {
 		dt_unit_t *tu = s->handlers[proto].handler.unit;
 		for (unsigned i = 0; i < tu->size; ++i) {
-			ref_retain((ref_t *)newlist);
-			s->handlers[proto].handler.thread_state[i] |= ServerReload;
 			s->handlers[proto].handler.thread_id[i] = thread_count++;
-			if (s->state & ServerRunning) {
-				dt_activate(tu->threads[i]);
-				dt_signalize(tu->threads[i], SIGALRM);
-			}
 		}
 	}
 
-	ref_release(&oldlist->ref);
-
-	return bound;
+	return KNOT_EOK;
 }
 
 int server_init(server_t *server, int bg_workers)
@@ -417,28 +416,18 @@ int server_init(server_t *server, int bg_workers)
 		return KNOT_ENOMEM;
 	}
 
-	char *journal_dir = conf_journalfile(conf());
-	conf_val_t journal_size = conf_default_get(conf(), C_MAX_JOURNAL_DB_SIZE);
-	conf_val_t journal_mode = conf_default_get(conf(), C_JOURNAL_DB_MODE);
-	int ret = journal_db_init(&server->journal_db, journal_dir,
-	                          conf_int(&journal_size), conf_opt(&journal_mode));
+	char *journal_dir = conf_db(conf(), C_JOURNAL_DB);
+	conf_val_t journal_size = conf_db_param(conf(), C_JOURNAL_DB_MAX_SIZE, C_MAX_JOURNAL_DB_SIZE);
+	conf_val_t journal_mode = conf_db_param(conf(), C_JOURNAL_DB_MODE, C_JOURNAL_DB_MODE);
+	knot_lmdb_init(&server->journaldb, journal_dir, conf_int(&journal_size), journal_env_flags(conf_opt(&journal_mode), false), NULL);
 	free(journal_dir);
-	if (ret != KNOT_EOK) {
-		worker_pool_destroy(server->workers);
-		evsched_deinit(&server->sched);
-		return ret;
-	}
 
-	char *kasp_dir = conf_kaspdir(conf());
-	conf_val_t kasp_size = conf_default_get(conf(), C_MAX_KASP_DB_SIZE);
-	ret = kasp_db_init(kaspdb(), kasp_dir, conf_int(&kasp_size));
-	free(kasp_dir);
-	if (ret != KNOT_EOK) {
-		journal_db_close(&server->journal_db);
-		worker_pool_destroy(server->workers);
-		evsched_deinit(&server->sched);
-		return ret;
-	}
+	kasp_db_ensure_init(&server->kaspdb, conf());
+
+	char *timer_dir = conf_db(conf(), C_TIMER_DB);
+	conf_val_t timer_size = conf_db_param(conf(), C_TIMER_DB_MAX_SIZE, C_MAX_TIMER_DB_SIZE);
+	knot_lmdb_init(&server->timerdb, timer_dir, conf_int(&timer_size), 0, NULL);
+	free(timer_dir);
 
 	return KNOT_EOK;
 }
@@ -449,35 +438,36 @@ void server_deinit(server_t *server)
 		return;
 	}
 
-	/* Free remaining interfaces. */
-	if (server->ifaces) {
-		iface_t *n = NULL, *m = NULL;
-		WALK_LIST_DELSAFE(n, m, server->ifaces->l) {
-			server_remove_iface(n);
+	/* Save zone timers. */
+	if (server->zone_db != NULL) {
+		log_info("updating persistent timer DB");
+		int ret = zone_timers_write_all(&server->timerdb, server->zone_db);
+		if (ret != KNOT_EOK) {
+			log_warning("failed to update persistent timer DB (%s)",
+				    knot_strerror(ret));
 		}
-		free(server->ifaces);
 	}
 
+	/* Free remaining interfaces. */
+	server_deinit_iface_list(server->ifaces);
+
 	/* Free threads and event handlers. */
 	worker_pool_destroy(server->workers);
 
 	/* Free zone database. */
-	knot_zonedb_deep_free(&server->zone_db);
+	knot_zonedb_deep_free(&server->zone_db, true);
 
 	/* Free remaining events. */
 	evsched_deinit(&server->sched);
 
+	/* Close persistent timers DB. */
+	knot_lmdb_deinit(&server->timerdb);
+
 	/* Close kasp_db. */
-	kasp_db_close(kaspdb());
+	knot_lmdb_deinit(&server->kaspdb);
 
 	/* Close journal database if open. */
-	journal_db_close(&server->journal_db);
-
-	/* Close persistent timers database. */
-	zone_timers_close(server->timers_db);
-
-	/* Clear the structure. */
-	memset(server, 0, sizeof(server_t));
+	knot_lmdb_deinit(&server->journaldb);
 }
 
 static int server_init_handler(server_t *server, int index, int thread_count,
@@ -524,7 +514,6 @@ static void server_free_handler(iohandler_t *h)
 	dt_delete(&h->unit);
 	free(h->thread_state);
 	free(h->thread_id);
-	memset(h, 0, sizeof(iohandler_t));
 }
 
 int server_start(server_t *server, bool async)
@@ -590,14 +579,15 @@ static int reload_conf(conf_t *new_conf)
 		log_info("reloading configuration file '%s'", filename);
 
 		/* Import the configuration file. */
-		ret = conf_import(new_conf, filename, true);
+		ret = conf_import(new_conf, filename, true, false);
 		if (ret != KNOT_EOK) {
 			log_error("failed to load configuration file (%s)",
 			          knot_strerror(ret));
 			return ret;
 		}
 	} else {
-		log_info("reloading configuration database");
+		log_info("reloading configuration database '%s'",
+		         knot_db_lmdb_get_path(new_conf->db));
 
 		/* Re-load extra modules. */
 		for (conf_iter_t iter = conf_iter(new_conf, C_MODULE);
@@ -626,6 +616,83 @@ static int reload_conf(conf_t *new_conf)
 	return KNOT_EOK;
 }
 
+/*! \brief Check if parameter listen has been changed since knotd started. */
+static bool listen_changed(conf_t *conf, server_t *server)
+{
+	assert(server->ifaces);
+
+	conf_val_t listen_val = conf_get(conf, C_SRV, C_LISTEN);
+	size_t new_count = conf_val_count(&listen_val);
+	size_t old_count = list_size(server->ifaces);
+	if (new_count != old_count) {
+		return true;
+	}
+
+	conf_val_t rundir_val = conf_get(conf, C_SRV, C_RUNDIR);
+	char *rundir = conf_abs_path(&rundir_val, NULL);
+	size_t matches = 0;
+
+	/* Find matching interfaces. */
+	while (listen_val.code == KNOT_EOK) {
+		struct sockaddr_storage addr = conf_addr(&listen_val, rundir);
+		bool found = false;
+		iface_t *iface;
+		WALK_LIST(iface, *server->ifaces) {
+			if (sockaddr_cmp(&addr, &iface->addr, false) == 0) {
+				matches++;
+				found = true;
+				break;
+			}
+		}
+
+		if (!found) {
+			break;
+		}
+		conf_val_next(&listen_val);
+	}
+
+	free(rundir);
+
+	return matches != old_count;
+}
+
+/*! \brief Log warnings if config change requires a restart. */
+static void warn_server_reconfigure(conf_t *conf, server_t *server)
+{
+	const char *msg = "changes of %s require restart to take effect";
+
+	static bool warn_tcp_reuseport = true;
+	static bool warn_udp = true;
+	static bool warn_tcp = true;
+	static bool warn_bg = true;
+	static bool warn_listen = true;
+
+	if (warn_tcp_reuseport && conf->cache.srv_tcp_reuseport != conf_tcp_reuseport(conf)) {
+		log_warning(msg, "tcp-reuseport");
+		warn_tcp_reuseport = false;
+	}
+
+	if (warn_udp && server->handlers[IO_UDP].size != conf_udp_threads(conf)) {
+		log_warning(msg, "udp-workers");
+		warn_udp = false;
+	}
+
+	if (warn_tcp && server->handlers[IO_TCP].size != conf_tcp_threads(conf)) {
+		log_warning(msg, "tcp-workers");
+		warn_tcp = false;
+	}
+
+	if (warn_bg && conf->cache.srv_bg_threads != conf_bg_threads(conf)) {
+		log_warning(msg, "background-workers");
+		warn_bg = false;
+	}
+
+	if (warn_listen && listen_changed(conf, server)) {
+		log_warning(msg, "listen");
+		warn_listen = false;
+	}
+}
+
 int server_reload(server_t *server)
 {
 	if (server == NULL) {
@@ -658,7 +725,7 @@ int server_reload(server_t *server)
 			return ret;
 		}
 
-		conf_activate_modules(new_conf, NULL, new_conf->query_modules,
+		conf_activate_modules(new_conf, server, NULL, new_conf->query_modules,
 		                      &new_conf->query_plan);
 	}
 
@@ -679,6 +746,7 @@ int server_reload(server_t *server)
 	}
 	if (full || (flags & CONF_IO_FRLD_SRV)) {
 		server_reconfigure(conf(), server);
+		warn_server_reconfigure(conf(), server);
 		stats_reconfigure(conf(), server);
 	}
 	if (full || (flags & (CONF_IO_FRLD_ZONES | CONF_IO_FRLD_ZONE))) {
@@ -714,100 +782,64 @@ void server_stop(server_t *server)
 	server->state &= ~ServerRunning;
 }
 
-static int reset_handler(server_t *server, int index, unsigned size, runnable_t run)
+static int set_handler(server_t *server, int index, unsigned size, runnable_t run)
 {
-	if (server->handlers[index].size != size) {
-		/* Free old handlers */
-		if (server->handlers[index].size > 0) {
-			server_free_handler(&server->handlers[index].handler);
-		}
-
-		/* Initialize I/O handlers. */
-		int ret = server_init_handler(server, index, size, run, NULL);
-		if (ret != KNOT_EOK) {
-			return ret;
-		}
-
-		/* Start if server is running. */
-		if (server->state & ServerRunning) {
-			ret = dt_start(server->handlers[index].handler.unit);
-			if (ret != KNOT_EOK) {
-				return ret;
-			}
-		}
-		server->handlers[index].size = size;
+	/* Initialize I/O handlers. */
+	int ret = server_init_handler(server, index, size, run, NULL);
+	if (ret != KNOT_EOK) {
+		return ret;
 	}
 
+	server->handlers[index].size = size;
+
 	return KNOT_EOK;
 }
 
 /*! \brief Reconfigure UDP and TCP query processing threads. */
-static int reconfigure_threads(conf_t *conf, server_t *server)
+static int configure_threads(conf_t *conf, server_t *server)
 {
-	int ret = reset_handler(server, IO_UDP, conf_udp_threads(conf), udp_master);
+	int ret = set_handler(server, IO_UDP, conf->cache.srv_udp_threads, udp_master);
 	if (ret != KNOT_EOK) {
 		return ret;
 	}
 
-	return reset_handler(server, IO_TCP, conf_tcp_threads(conf), tcp_master);
+	return set_handler(server, IO_TCP, conf->cache.srv_tcp_threads, tcp_master);
 }
 
 static int reconfigure_journal_db(conf_t *conf, server_t *server)
 {
-	char *journal_dir = conf_journalfile(conf);
-	conf_val_t journal_size = conf_default_get(conf, C_MAX_JOURNAL_DB_SIZE);
-	conf_val_t journal_mode = conf_default_get(conf, C_JOURNAL_DB_MODE);
-	bool changed_path = (strcmp(journal_dir, server->journal_db->path) != 0);
-	bool changed_size = (conf_int(&journal_size) != server->journal_db->fslimit);
-	bool changed_mode = (conf_opt(&journal_mode) != server->journal_db->mode);
-	int ret = KNOT_EOK;
-
-	if (server->journal_db->db != NULL) {
-		if (changed_path) {
-			log_warning("ignored reconfiguration of journal DB path (already open)");
-		}
-		if (changed_size) {
-			log_warning("ignored reconfiguration of journal DB max size (already open)");
-		}
-		if (changed_mode) {
-			log_warning("ignored reconfiguration of journal DB mode (already open)");
-		}
-	} else if (changed_path || changed_size || changed_mode) {
-		journal_db_t *newjdb = NULL;
-		ret = journal_db_init(&newjdb, journal_dir, conf_int(&journal_size),
-		                      conf_opt(&journal_mode));
-		if (ret == KNOT_EOK) {
-			journal_db_close(&server->journal_db);
-			server->journal_db = newjdb;
-		}
+	char *journal_dir = conf_db(conf, C_JOURNAL_DB);
+	conf_val_t journal_size = conf_db_param(conf, C_JOURNAL_DB_MAX_SIZE, C_MAX_JOURNAL_DB_SIZE);
+	conf_val_t journal_mode = conf_db_param(conf, C_JOURNAL_DB_MODE, C_JOURNAL_DB_MODE);
+	int ret = knot_lmdb_reinit(&server->journaldb, journal_dir, conf_int(&journal_size),
+	                           journal_env_flags(conf_opt(&journal_mode), false));
+	if (ret != KNOT_EOK) {
+		log_warning("ignored reconfiguration of journal DB (%s)", knot_strerror(ret));
 	}
 	free(journal_dir);
 
-	return ret;
+	return KNOT_EOK; // not "ret"
 }
 
 static int reconfigure_kasp_db(conf_t *conf, server_t *server)
 {
-	char *kasp_dir = conf_kaspdir(conf);
-	conf_val_t kasp_size = conf_default_get(conf, C_MAX_KASP_DB_SIZE);
-	int ret = kasp_db_reconfigure(kaspdb(), kasp_dir, conf_int(&kasp_size));
-	switch (ret) {
-	case KNOT_EBUSY:
-		log_warning("ignored reconfiguration of KASP DB path (already open)");
-		break;
-	case KNOT_EEXIST:
-		ret = KNOT_EBUSY;
-		log_warning("ignored reconfiguration of KASP DB max size (already open)");
-		break;
-	case KNOT_ENODIFF:
-	case KNOT_EOK:
-		ret = KNOT_EOK;
-		break;
-	default:
-		break;
+	char *kasp_dir = conf_db(conf, C_KASP_DB);
+	conf_val_t kasp_size = conf_db_param(conf, C_KASP_DB_MAX_SIZE, C_MAX_KASP_DB_SIZE);
+	int ret = knot_lmdb_reinit(&server->kaspdb, kasp_dir, conf_int(&kasp_size), 0);
+	if (ret != KNOT_EOK) {
+		log_warning("ignored reconfiguration of KASP DB (%s)", knot_strerror(ret));
 	}
 	free(kasp_dir);
 
+	return KNOT_EOK; // not "ret"
+}
+
+static int reconfigure_timer_db(conf_t *conf, server_t *server)
+{
+	char *timer_dir = conf_db(conf, C_TIMER_DB);
+	conf_val_t timer_size = conf_db_param(conf, C_TIMER_DB_MAX_SIZE, C_MAX_TIMER_DB_SIZE);
+	int ret = knot_lmdb_reconfigure(&server->timerdb, timer_dir, conf_int(&timer_size), 0);
+	free(timer_dir);
 	return ret;
 }
 
@@ -817,59 +849,52 @@ void server_reconfigure(conf_t *conf, server_t *server)
 		return;
 	}
 
+	int ret;
+
 	/* First reconfiguration. */
 	if (!(server->state & ServerRunning)) {
 		log_info("Knot DNS %s starting", PACKAGE_VERSION);
-	}
 
-	/* Reconfigure server threads. */
-	int ret;
-	if ((ret = reconfigure_threads(conf, server)) < 0) {
-		log_error("failed to reconfigure server threads (%s)",
-		          knot_strerror(ret));
+		if (conf->filename != NULL) {
+			log_info("loaded configuration file '%s'",
+			         conf->filename);
+		} else {
+			log_info("loaded configuration database '%s'",
+			         knot_db_lmdb_get_path(conf->db));
+		}
+
+		/* Configure server threads. */
+		if ((ret = configure_threads(conf, server)) != KNOT_EOK) {
+			log_error("failed to configure server threads (%s)",
+			          knot_strerror(ret));
+		}
+
+		/* Configure sockets. */
+		if ((ret = configure_sockets(conf, server)) != KNOT_EOK) {
+			log_error("failed to configure server sockets (%s)",
+			          knot_strerror(ret));
+		}
 	}
 
 	/* Reconfigure journal DB. */
-	if ((ret = reconfigure_journal_db(conf, server)) < 0) {
+	if ((ret = reconfigure_journal_db(conf, server)) != KNOT_EOK) {
 		log_error("failed to reconfigure journal DB (%s)",
 		          knot_strerror(ret));
 	}
 
 	/* Reconfigure KASP DB. */
-	if ((ret = reconfigure_kasp_db(conf, server)) < 0) {
+	if ((ret = reconfigure_kasp_db(conf, server)) != KNOT_EOK) {
 		log_error("failed to reconfigure KASP DB (%s)",
 		          knot_strerror(ret));
 	}
 
-	/* Update bound sockets. */
-	if ((ret = reconfigure_sockets(conf, server)) < 0) {
-		log_error("failed to reconfigure server sockets (%s)",
+	/* Reconfiure Timer DB. */
+	if ((ret = reconfigure_timer_db(conf, server)) != KNOT_EOK) {
+		log_error("failed to reconfigure Timer DB (%s)",
 		          knot_strerror(ret));
 	}
 }
 
-static void reopen_timers_database(conf_t *conf, server_t *server)
-{
-	zone_timers_close(server->timers_db);
-	server->timers_db = NULL;
-
-	conf_val_t val = conf_default_get(conf, C_STORAGE);
-	char *storage = conf_abs_path(&val, NULL);
-	val = conf_default_get(conf, C_TIMER_DB);
-	char *timer_db = conf_abs_path(&val, storage);
-	free(storage);
-	val = conf_default_get(conf, C_MAX_TIMER_DB_SIZE);
-	size_t mapsize = conf_int(&val);
-
-	int ret = zone_timers_open(timer_db, &server->timers_db, mapsize);
-	if (ret != KNOT_EOK) {
-		log_warning("cannot open persistent timer DB '%s' (%s)",
-		            timer_db, knot_strerror(ret));
-	}
-
-	free(timer_db);
-}
-
 void server_update_zones(conf_t *conf, server_t *server)
 {
 	if (conf == NULL || server == NULL) {
@@ -886,7 +911,6 @@ void server_update_zones(conf_t *conf, server_t *server)
 	worker_pool_wait(server->workers);
 
 	/* Reload zone database and free old zones. */
-	reopen_timers_database(conf, server);
 	zonedb_reload(conf, server);
 
 	/* Trim extra heap. */
@@ -898,35 +922,3 @@ void server_update_zones(conf_t *conf, server_t *server)
 		knot_zonedb_foreach(server->zone_db, zone_events_start);
 	}
 }
-
-ref_t *server_set_ifaces(server_t *server, fdset_t *fds, int index, int thread_id)
-{
-	if (server == NULL || server->ifaces == NULL || fds == NULL) {
-		return NULL;
-	}
-
-	rcu_read_lock();
-	fdset_clear(fds);
-
-	iface_t *i = NULL;
-	WALK_LIST(i, server->ifaces->l) {
-#ifdef ENABLE_REUSEPORT
-		int udp_id = thread_id % i->fd_udp_count;
-#else
-		int udp_id = 0;
-#endif
-		switch(index) {
-		case IO_TCP:
-			fdset_add(fds, i->fd_tcp, POLLIN, NULL);
-			break;
-		case IO_UDP:
-			fdset_add(fds, i->fd_udp[udp_id], POLLIN, NULL);
-			break;
-		default:
-			assert(0);
-		}
-	}
-	rcu_read_unlock();
-
-	return &server->ifaces->ref;
-}
