@@ -1,4 +1,4 @@
-/*  Copyright (C) 2018 CZ.NIC, z.s.p.o. <knot-dns@labs.nic.cz>
+/*  Copyright (C) 2019 CZ.NIC, z.s.p.o. <knot-dns@labs.nic.cz>
 
     This program is free software: you can redistribute it and/or modify
     it under the terms of the GNU General Public License as published by
@@ -11,7 +11,7 @@
     GNU General Public License for more details.
 
     You should have received a copy of the GNU General Public License
-    along with this program.  If not, see <http://www.gnu.org/licenses/>.
+    along with this program.  If not, see <https://www.gnu.org/licenses/>.
  */
 
 #pragma once
@@ -21,8 +21,8 @@
 #include "knot/conf/conf.h"
 #include "knot/common/evsched.h"
 #include "knot/common/fdset.h"
+#include "knot/journal/knot_lmdb.h"
 #include "knot/server/dthreads.h"
-#include "knot/common/ref.h"
 #include "knot/worker/pool.h"
 #include "knot/zone/zonedb.h"
 #include "contrib/ucw/lists.h"
@@ -45,7 +45,6 @@ typedef struct iohandler {
 typedef enum {
 	ServerIdle    = 0 << 0, /*!< Server is idle. */
 	ServerRunning = 1 << 0, /*!< Server is running. */
-	ServerReload  = 1 << 1  /*!< Server reload requested. */
 } server_state;
 
 /*!
@@ -55,7 +54,8 @@ typedef struct iface {
 	struct node n;
 	int *fd_udp;
 	int fd_udp_count;
-	int fd_tcp;
+	int *fd_tcp;
+	int fd_tcp_count;
 	struct sockaddr_storage addr;
 } iface_t;
 
@@ -65,12 +65,6 @@ enum {
 	IO_TCP = 1
 };
 
-typedef struct ifacelist {
-	ref_t ref;
-	list_t l;
-	list_t u;
-} ifacelist_t;
-
 /*!
  * \brief Main server structure.
  *
@@ -83,8 +77,9 @@ typedef struct server {
 
 	/*! \brief Zone database. */
 	knot_zonedb_t *zone_db;
-	knot_db_t *timers_db;
-	journal_db_t *journal_db;
+	knot_lmdb_db_t timerdb;
+	knot_lmdb_db_t journaldb;
+	knot_lmdb_db_t kaspdb;
 
 	/*! \brief I/O handlers. */
 	struct {
@@ -99,7 +94,7 @@ typedef struct server {
 	evsched_t sched;
 
 	/*! \brief List of interfaces. */
-	ifacelist_t *ifaces;
+	list_t *ifaces;
 
 } server_t;
 
@@ -167,15 +162,3 @@ void server_reconfigure(conf_t *conf, server_t *server);
  * Routine for dynamic server zones reconfiguration.
  */
 void server_update_zones(conf_t *conf, server_t *server);
-
-/*!
- * \brief Update fdsets from current interfaces list.
- *
- * \param server    Server.
- * \param fds       File descriptor set.
- * \param index     I/O index (UDP/TCP).
- * \param thread_id Thread ID used for geting UDP ID.
- *
- * \return new interface list
- */
-ref_t *server_set_ifaces(server_t *server, fdset_t *fds, int index, int thread_id);
