@@ -1,4 +1,4 @@
-/*  Copyright (C) 2018 CZ.NIC, z.s.p.o. <knot-dns@labs.nic.cz>
+/*  Copyright (C) 2019 CZ.NIC, z.s.p.o. <knot-dns@labs.nic.cz>
 
     This program is free software: you can redistribute it and/or modify
     it under the terms of the GNU General Public License as published by
@@ -11,7 +11,7 @@
     GNU General Public License for more details.
 
     You should have received a copy of the GNU General Public License
-    along with this program.  If not, see <http://www.gnu.org/licenses/>.
+    along with this program.  If not, see <https://www.gnu.org/licenses/>.
  */
 
 #include <assert.h>
@@ -45,7 +45,7 @@ static int add_rr_to_list(list_t *l, const knot_rrset_t *rr)
 	WALK_LIST(n, *l) {
 		ptrnode_t *ptr_n = (ptrnode_t *)n;
 		knot_rrset_t *rrset = (knot_rrset_t *)ptr_n->d;
-		if (knot_rrset_equal(rr, rrset, KNOT_RRSET_COMPARE_HEADER)) {
+		if (rrset->type == rr->type && knot_dname_is_equal(rrset->owner, rr->owner)) {
 			return knot_rdataset_merge(&rrset->rrs, &rr->rrs, NULL);
 		}
 	};
@@ -70,7 +70,7 @@ static int check_rrset_exists(zone_update_t *update, const knot_rrset_t *rrset,
 	} else {
 		knot_rrset_t found = node_rrset(node, rrset->type);
 		assert(!knot_rrset_empty(&found));
-		if (knot_rrset_equal(&found, rrset, KNOT_RRSET_COMPARE_WHOLE)) {
+		if (knot_rrset_equal(&found, rrset, false)) {
 			return KNOT_EOK;
 		} else {
 			*rcode = KNOT_RCODE_NXRRSET;
@@ -272,6 +272,7 @@ static inline bool is_node_removal(const knot_rrset_t *rr)
 static bool should_replace(const knot_rrset_t *rrset)
 {
 	return rrset->type == KNOT_RRTYPE_CNAME ||
+	       rrset->type == KNOT_RRTYPE_DNAME ||
 	       rrset->type == KNOT_RRTYPE_NSEC3PARAM;
 }
 
@@ -370,7 +371,7 @@ static int process_add_cname(const zone_node_t *node,
 	knot_rrset_t cname = node_rrset(node, KNOT_RRTYPE_CNAME);
 	if (!knot_rrset_empty(&cname)) {
 		// If they are identical, ignore.
-		if (knot_rrset_equal(&cname, rr, KNOT_RRSET_COMPARE_WHOLE)) {
+		if (knot_rrset_equal(&cname, rr, true)) {
 			return KNOT_EOK;
 		}
 
@@ -429,7 +430,7 @@ static int process_add_soa(const zone_node_t *node,
 
 	// Get current SOA RR.
 	knot_rrset_t removed = node_rrset(node, KNOT_RRTYPE_SOA);
-	if (knot_rrset_equal(&removed, rr, KNOT_RRSET_COMPARE_WHOLE)) {
+	if (knot_rrset_equal(&removed, rr, true)) {
 		// If they are identical, ignore.
 		return KNOT_EOK;
 	}
@@ -511,7 +512,10 @@ static int process_rem_rr(const knot_rrset_t *rr,
 		return KNOT_EOK;
 	}
 
-	return zone_update_remove(update, rr);
+	knot_rrset_t rr_ttl = *rr;
+	rr_ttl.ttl = to_modify.ttl;
+
+	return zone_update_remove(update, &rr_ttl);
 }
 
 /*!< \brief Removes RRSet from zone. */
@@ -553,24 +557,16 @@ static int process_rem_node(const knot_rrset_t *rr,
 		return KNOT_EOK;
 	}
 
-	zone_node_t *node_copy = node_shallow_copy(node, NULL);
-	if (node_copy == NULL) {
-		return KNOT_ENOMEM;
-	}
-
 	// Remove all RRSets from node
-	size_t rrset_count = node_copy->rrset_count;
+	size_t rrset_count = node->rrset_count;
 	for (int i = 0; i < rrset_count; ++i) {
-		knot_rrset_t rrset = node_rrset_at(node_copy, rrset_count - i - 1);
-		int ret = process_rem_rrset(&rrset, node_copy, update);
+		knot_rrset_t rrset = node_rrset_at(node, rrset_count - i - 1);
+		int ret = process_rem_rrset(&rrset, node, update);
 		if (ret != KNOT_EOK) {
-			node_free(node_copy, NULL);
 			return ret;
 		}
 	}
 
-	node_free(node_copy, NULL);
-
 	return KNOT_EOK;
 }
 
