@@ -11,7 +11,7 @@
     GNU General Public License for more details.
 
     You should have received a copy of the GNU General Public License
-    along with this program.  If not, see <http://www.gnu.org/licenses/>.
+    along with this program.  If not, see <https://www.gnu.org/licenses/>.
  */
 
 #pragma once
@@ -79,7 +79,7 @@ zone_contents_t *zonefile_load(zloader_t *loader);
  *
  * \return KNOT_E*
  */
-int zonefile_exists(const char *path, time_t *mtime);
+int zonefile_exists(const char *path, struct timespec *mtime);
 
 /*!
  * \brief Write zone contents to zone file.
