@@ -1,4 +1,4 @@
-/*  Copyright (C) 2018 CZ.NIC, z.s.p.o. <knot-dns@labs.nic.cz>
+/*  Copyright (C) 2019 CZ.NIC, z.s.p.o. <knot-dns@labs.nic.cz>
 
     This program is free software: you can redistribute it and/or modify
     it under the terms of the GNU General Public License as published by
@@ -11,7 +11,7 @@
     GNU General Public License for more details.
 
     You should have received a copy of the GNU General Public License
-    along with this program.  If not, see <http://www.gnu.org/licenses/>.
+    along with this program.  If not, see <https://www.gnu.org/licenses/>.
  */
 
 #include <assert.h>
@@ -20,7 +20,6 @@
 
 #include "libknot/attribute.h"
 #include "libknot/rdataset.h"
-#include "libknot/errcode.h"
 #include "contrib/mempattern.h"
 
 static knot_rdata_t *rr_seek(const knot_rdataset_t *rrs, uint16_t pos)
@@ -50,51 +49,39 @@ static int find_rr_pos(const knot_rdataset_t *rrs, const knot_rdata_t *rr)
 	return KNOT_ENOENT;
 }
 
-static int add_rr_at(knot_rdataset_t *rrs, const knot_rdata_t *rr, uint16_t pos,
+static int add_rr_at(knot_rdataset_t *rrs, const knot_rdata_t *rr, knot_rdata_t *ins_pos,
                      knot_mm_t *mm)
 {
 	assert(rrs);
 	assert(rr);
-	assert(pos <= rrs->count);
+	const size_t ins_offset = (uint8_t *)ins_pos - (uint8_t *)rrs->rdata;
+	assert(ins_offset <= rrs->size);
 
 	if (rrs->count == UINT16_MAX) {
 		return KNOT_ESPACE;
+	} else if (rrs->size > UINT32_MAX - knot_rdata_size(UINT16_MAX)) {
+		return KNOT_ESPACE;
 	}
 
-	size_t total_size = knot_rdataset_size(rrs);
-	size_t new_size = knot_rdata_size(rr->len);
+	const size_t rr_size = knot_rdata_size(rr->len);
 
 	// Realloc RDATA.
-	knot_rdata_t *tmp = mm_realloc(mm, rrs->rdata, total_size + new_size,
-	                               total_size);
+	knot_rdata_t *tmp = mm_realloc(mm, rrs->rdata, rrs->size + rr_size,
+	                               rrs->size);
 	if (tmp == NULL) {
 		return KNOT_ENOMEM;
 	} else {
 		rrs->rdata = tmp;
 	}
 
-	if (rrs->count == 0 || pos == rrs->count) {
-		// No need to rearange RDATA.
-		rrs->count++;
-		knot_rdata_t *new_rr = rr_seek(rrs, pos);
-		knot_rdata_init(new_rr, rr->len, rr->data);
-		return KNOT_EOK;
-	}
-
-	// RDATA have to be rearanged.
-	knot_rdata_t *old_rr = rr_seek(rrs, pos);
-	knot_rdata_t *last_rr = rr_seek(rrs, rrs->count - 1);
-
-	// Make space for new RDATA by moving the array.
-	uint8_t *dst = (uint8_t *)old_rr + new_size;
-	assert(old_rr <= last_rr);
-	size_t len = ((uint8_t *)last_rr - (uint8_t *)old_rr) +
-	             knot_rdata_size(last_rr->len);
-	memmove(dst, old_rr, len);
+	uint8_t *ins_pos_raw = (uint8_t *)rrs->rdata + ins_offset;
+	// RDATA may have to be rearanged.  Moving zero-length region is OK.
+	memmove(ins_pos_raw + rr_size, ins_pos_raw, rrs->size - ins_offset);
 
 	// Set new RDATA.
-	knot_rdata_init(old_rr, rr->len, rr->data);
+	knot_rdata_init((knot_rdata_t *)ins_pos_raw, rr->len, rr->data);
 	rrs->count++;
+	rrs->size += rr_size;
 
 	return KNOT_EOK;
 }
@@ -108,7 +95,6 @@ static int remove_rr_at(knot_rdataset_t *rrs, uint16_t pos, knot_mm_t *mm)
 	knot_rdata_t *old_rr = rr_seek(rrs, pos);
 	knot_rdata_t *last_rr = rr_seek(rrs, rrs->count - 1);
 
-	size_t total_size = knot_rdataset_size(rrs);
 	size_t old_size = knot_rdata_size(old_rr->len);
 
 	// Move RDATA.
@@ -119,8 +105,8 @@ static int remove_rr_at(knot_rdataset_t *rrs, uint16_t pos, knot_mm_t *mm)
 
 	if (rrs->count > 1) {
 		// Realloc RDATA.
-		knot_rdata_t *tmp = mm_realloc(mm, rrs->rdata, total_size - old_size,
-		                               total_size);
+		knot_rdata_t *tmp = mm_realloc(mm, rrs->rdata, rrs->size - old_size,
+		                               rrs->size);
 		if (tmp == NULL) {
 			return KNOT_ENOMEM;
 		} else {
@@ -132,6 +118,7 @@ static int remove_rr_at(knot_rdataset_t *rrs, uint16_t pos, knot_mm_t *mm)
 		rrs->rdata = NULL;
 	}
 	rrs->count--;
+	rrs->size -= old_size;
 
 	return KNOT_EOK;
 }
@@ -155,13 +142,13 @@ int knot_rdataset_copy(knot_rdataset_t *dst, const knot_rdataset_t *src, knot_mm
 	}
 
 	dst->count = src->count;
-	size_t src_size = knot_rdataset_size(src);
-	dst->rdata = mm_alloc(mm, src_size);
+	dst->size = src->size;
+	dst->rdata = mm_alloc(mm, src->size);
 	if (dst->rdata == NULL) {
 		return KNOT_ENOMEM;
 	}
 
-	memcpy(dst->rdata, src->rdata, src_size);
+	memcpy(dst->rdata, src->rdata, src->size);
 
 	return KNOT_EOK;
 }
@@ -176,17 +163,6 @@ knot_rdata_t *knot_rdataset_at(const knot_rdataset_t *rrs, uint16_t pos)
 	return rr_seek(rrs, pos);
 }
 
-_public_
-size_t knot_rdataset_size(const knot_rdataset_t *rrs)
-{
-	if (rrs == NULL || rrs->count == 0) {
-		return 0;
-	}
-
-	const knot_rdata_t *last = rr_seek(rrs, rrs->count - 1);
-	return (uint8_t *)last + knot_rdata_size(last->len) - (uint8_t *)rrs->rdata;
-}
-
 _public_
 int knot_rdataset_add(knot_rdataset_t *rrs, const knot_rdata_t *rr, knot_mm_t *mm)
 {
@@ -194,60 +170,31 @@ int knot_rdataset_add(knot_rdataset_t *rrs, const knot_rdata_t *rr, knot_mm_t *m
 		return KNOT_EINVAL;
 	}
 
-	// First insert to empty rdataset.
-	if (rrs->count == 0) {
-		return add_rr_at(rrs, rr, 0, mm);
-	}
-
-	for (int i = rrs->count - 1; i >= 0; --i) {
-		const knot_rdata_t *rrset_rr = rr_seek(rrs, i);
-		int cmp = knot_rdata_cmp(rrset_rr, rr);
-		if (cmp == 0) {
-			// Duplicate - no need to add this RR.
-			return KNOT_EOK;
-		} else if (cmp < 0) {
-			// Found position to insert.
-			return add_rr_at(rrs, rr, i + 1, mm);
+	// Optimize a little for insertion at the end, for larger RRsets.
+	if (rrs->count > 4) {
+		knot_rdata_t *last = rr_seek(rrs, rrs->count - 1);
+		if (knot_rdata_cmp(last, rr) < 0) {
+			return add_rr_at(rrs, rr, knot_rdataset_next(last), mm);
 		}
 	}
 
-	// If flow gets here, it means that we should insert at the first position.
-	return add_rr_at(rrs, rr, 0, mm);
-}
-
-_public_
-int knot_rdataset_reserve(knot_rdataset_t *rrs, uint16_t size, knot_mm_t *mm)
-{
-	if (rrs == NULL) {
-		return KNOT_EINVAL;
-	} else if (rrs->count == UINT16_MAX) {
-		return KNOT_ESPACE;
+	// Look for the right place to insert.
+	knot_rdata_t *ins_pos = rrs->rdata;
+	for (int i = 0; i < rrs->count; ++i, ins_pos = knot_rdataset_next(ins_pos)) {
+		int cmp = knot_rdata_cmp(ins_pos, rr);
+		if (cmp == 0) {
+			// Duplicate - no need to add this RR.
+			return KNOT_EOK;
+		} else if (cmp > 0) {
+			// Found position to insert.
+			return add_rr_at(rrs, rr, ins_pos, mm);
+		}
 	}
 
-	size_t old_size = knot_rdataset_size(rrs);
-	size_t new_size = old_size + knot_rdata_size(size);
+	assert((uint8_t *)rrs->rdata + rrs->size == (uint8_t *)ins_pos);
 
-	knot_rdata_t *tmp = mm_realloc(mm, rrs->rdata, new_size, old_size);
-	if (tmp == NULL) {
-		return KNOT_ENOMEM;
-	}
-	rrs->rdata = tmp;
-	rrs->count++;
-
-	// We have to initialise the 'size' field in the reserved space.
-	rr_seek(rrs, rrs->count - 1)->len = size;
-
-	return KNOT_EOK;
-}
-
-_public_
-int knot_rdataset_unreserve(knot_rdataset_t *rrs, knot_mm_t *mm)
-{
-	if (rrs == NULL || rrs->count == 0) {
-		return KNOT_EINVAL;
-	}
-
-	return remove_rr_at(rrs, rrs->count - 1, mm);
+	// If flow gets here, it means that we should insert at the current position (append).
+	return add_rr_at(rrs, rr, ins_pos, mm);
 }
 
 _public_
