@@ -1,4 +1,4 @@
-/*  Copyright (C) 2018 CZ.NIC, z.s.p.o. <knot-dns@labs.nic.cz>
+/*  Copyright (C) 2019 CZ.NIC, z.s.p.o. <knot-dns@labs.nic.cz>
 
     This program is free software: you can redistribute it and/or modify
     it under the terms of the GNU General Public License as published by
@@ -11,7 +11,7 @@
     GNU General Public License for more details.
 
     You should have received a copy of the GNU General Public License
-    along with this program.  If not, see <http://www.gnu.org/licenses/>.
+    along with this program.  If not, see <https://www.gnu.org/licenses/>.
  */
 
 #include <tap/basic.h>
@@ -31,7 +31,7 @@
 
 char *format_key(conf_io_t *io)
 {
-	char id[KNOT_DNAME_TXT_MAXLEN + 1] = "\0";
+	knot_dname_txt_storage_t id;
 	size_t id_len = sizeof(id);
 
 	// Get the textual item id.
@@ -209,11 +209,12 @@ static void test_conf_io_abort(void)
 	SKIP_OPENBSD
 #else
 	// Test child persistence after subchild abort.
-
-	ok(conf_io_begin(false) == KNOT_EOK, "begin parent txn");
-	char idx[2] = { '0' };
-	ok(conf_io_set("server", "version", NULL, idx) ==
-	   KNOT_EOK, "set single value '%s'", idx);
+	{
+		char idx[2] = { '0' };
+		ok(conf_io_begin(false) == KNOT_EOK, "begin parent txn");
+		ok(conf_io_set("server", "version", NULL, idx) ==
+		   KNOT_EOK, "set single value '%s'", idx);
+	}
 
 	for (int i = 1; i < CONF_MAX_TXN_DEPTH; i++) {
 		char idx[2] = { '0' + i };
@@ -265,11 +266,12 @@ static void test_conf_io_commit(void)
 	SKIP_OPENBSD
 #else
 	// Test subchild persistence after commit.
-
-	ok(conf_io_begin(false) == KNOT_EOK, "begin parent txn");
-	char idx[2] = { '0' };
-	ok(conf_io_set("server", "version", NULL, idx) ==
-	   KNOT_EOK, "set single value '%s'", idx);
+	{
+		char idx[2] = { '0' };
+		ok(conf_io_begin(false) == KNOT_EOK, "begin parent txn");
+		ok(conf_io_set("server", "version", NULL, idx) ==
+		   KNOT_EOK, "set single value '%s'", idx);
+	}
 
 	for (int i = 1; i < CONF_MAX_TXN_DEPTH; i++) {
 		char idx[2] = { '0' + i };
@@ -294,7 +296,7 @@ static void test_conf_io_commit(void)
 	ok(conf_io_begin(false) == KNOT_EOK, "begin new parent txn");
 	conf_val_t val = conf_get_txn(conf(), conf()->io.txn, C_SERVER, C_VERSION);
 	ok(val.code == KNOT_EOK, "check entry");
-	idx[0] = '0' + CONF_MAX_TXN_DEPTH - 1;
+	char idx[2] = { '0' + CONF_MAX_TXN_DEPTH - 1 };
 	const char *data = conf_str(&val);
 	ok(strcmp(data, idx) == 0, "compare final data '%s'", data);
 	conf_io_abort(false);
@@ -900,34 +902,45 @@ static void test_conf_io_list(void)
 	ok(conf_io_list("server", &io) ==
 	   KNOT_EOK, "list group");
 	ref = "server.version\n"
-	      "server.background-workers\n"
 	      "server.listen\n"
-	      "server.tcp-handshake-timeout\n"
 	      "server.tcp-idle-timeout\n"
-	      "server.tcp-reply-timeout\n"
-	      "server.max-tcp-clients\n"
-	      "server.max-udp-payload\n"
-	      "server.max-ipv4-udp-payload\n"
-	      "server.max-ipv6-udp-payload\n"
+	      "server.tcp-io-timeout\n"
+	      "server.tcp-remote-io-timeout\n"
+	      "server.tcp-max-clients\n"
+	      "server.tcp-reuseport\n"
+	      "server.udp-workers\n"
+	      "server.tcp-workers\n"
+	      "server.background-workers\n"
+	      "server.udp-max-payload\n"
+	      "server.udp-max-payload-ipv4\n"
+	      "server.udp-max-payload-ipv6\n"
 	      "server.edns-client-subnet\n"
-	      "server.answer-rotation";
+	      "server.answer-rotation\n"
+	      "server.max-tcp-clients\n"
+	      "server.max-udp-payload";
 	ok(strcmp(ref, out) == 0, "compare result");
 }
 
 static const yp_item_t desc_server[] = {
 	{ C_VERSION,              YP_TSTR,  YP_VNONE },
-	{ C_BG_WORKERS,           YP_TINT,  YP_VNONE },
 	{ C_LISTEN,               YP_TADDR, YP_VNONE, YP_FMULTI },
 	// Required config cache items - assert fix.
-	{ C_TCP_HSHAKE_TIMEOUT,   YP_TINT,  YP_VNONE },
 	{ C_TCP_IDLE_TIMEOUT,	  YP_TINT,  YP_VNONE },
-	{ C_TCP_REPLY_TIMEOUT,	  YP_TINT,  YP_VNONE },
-	{ C_MAX_TCP_CLIENTS,	  YP_TINT,  YP_VNONE },
-	{ C_MAX_UDP_PAYLOAD,      YP_TINT,  YP_VNONE },
-	{ C_MAX_IPV4_UDP_PAYLOAD, YP_TINT,  YP_VNONE },
-	{ C_MAX_IPV6_UDP_PAYLOAD, YP_TINT,  YP_VNONE },
+	{ C_TCP_IO_TIMEOUT,	  YP_TINT,  YP_VNONE },
+	{ C_TCP_RMT_IO_TIMEOUT,	  YP_TINT,  YP_VNONE },
+	{ C_TCP_MAX_CLIENTS,	  YP_TINT,  YP_VNONE },
+	{ C_TCP_REUSEPORT,	  YP_TBOOL, YP_VNONE },
+	{ C_UDP_WORKERS,	  YP_TINT,  YP_VNONE },
+	{ C_TCP_WORKERS,	  YP_TINT,  YP_VNONE },
+	{ C_BG_WORKERS,		  YP_TINT,  YP_VNONE },
+	{ C_UDP_MAX_PAYLOAD,      YP_TINT,  YP_VNONE },
+	{ C_UDP_MAX_PAYLOAD_IPV4, YP_TINT,  YP_VNONE },
+	{ C_UDP_MAX_PAYLOAD_IPV6, YP_TINT,  YP_VNONE },
 	{ C_ECS,                  YP_TBOOL, YP_VNONE },
 	{ C_ANS_ROTATION,         YP_TBOOL, YP_VNONE },
+	// Legacy items.
+	{ C_MAX_TCP_CLIENTS,      YP_TINT,  YP_VNONE },
+	{ C_MAX_UDP_PAYLOAD,      YP_TINT,  YP_VNONE },
 	{ NULL }
 };
 
